SystemOrganization addCategory: #Traits!
SystemOrganization addCategory: #'Traits-Composition'!
SystemOrganization addCategory: #'Traits-Kernel'!
SystemOrganization addCategory: #'Traits-Kernel-Traits'!

Trait named: #TApplyingOnClassSide
	uses: {}
	category: 'Traits-Kernel-Traits'!

!TApplyingOnClassSide methodsFor: 'composition' stamp: 'al 7/30/2004 09:07'!
assertConsistantCompositionsForNew: aTraitComposition
	"Applying or modifying a trait composition on the class side
	of a behavior has some restrictions."

	| baseTraits notAddable message |
	baseTraits := aTraitComposition traits select: [:each | each isBaseTrait].
	baseTraits isEmpty ifFalse: [
		notAddable := (baseTraits reject: [:each | each classSide methodDict isEmpty]).
		notAddable isEmpty ifFalse: [
			message := String streamContents: [:stream |
				stream nextPutAll: 'You can not add the base trait(s)'; cr.
				notAddable
					do: [:each | stream nextPutAll: each name]
					separatedBy: [ stream nextPutAll: ', '].
				stream cr; nextPutAll: 'to this composition because it/they define(s) methods on the class side.'].
		^TraitCompositionException signal: message]].
		
	(self instanceSide traitComposition traits asSet =
			(aTraitComposition traits
				select: [:each | each isClassTrait]
				thenCollect: [:each | each baseTrait]) asSet) ifFalse: [
				^TraitCompositionException signal: 'You can not add or remove class side traits on
				the class side of a composition. (But you can specify aliases or exclusions
				for existing traits or add a trait which does not have any methods on the class side.)']! !

!TApplyingOnClassSide methodsFor: 'composition' stamp: 'apb 8/24/2005 14:15'!
noteNewBaseTraitCompositionApplied: aTraitComposition
	"The argument is the new trait composition of my base trait - add
	the new traits or remove non existing traits on my class side composition.
	(Each class trait in my composition has its base trait on the instance side
	of the composition - manually added traits to the class side are always 
	base traits.)"
	
	| newComposition traitsFromInstanceSide |
	traitsFromInstanceSide := self traitComposition traits
		select: [:each | each isClassTrait]
		thenCollect: [:each | each baseTrait].
		
	newComposition := self traitComposition copyTraitExpression.
	(traitsFromInstanceSide copyWithoutAll: aTraitComposition traits) do: [:each |
		newComposition removeFromComposition: each classTrait].
	(aTraitComposition traits copyWithoutAll: traitsFromInstanceSide) do: [:each |
		newComposition add:  (each classTrait)].

	self setTraitComposition: newComposition! !

Trait named: #TBehaviorCategorization
	uses: {}
	category: 'Traits-Kernel-Traits'!

!TBehaviorCategorization methodsFor: 'organization' stamp: 'StephaneDucasse 10/11/2010 16:19'!
category
	"Answer the system organization category for the receiver. First check whether the
	category name stored in the ivar is still correct and only if this fails look it up
	(latter is much more expensive)"

	| result |
	self basicCategory ifNotNil: [ :symbol |
		((self environment organization listAtCategoryNamed: symbol) includes: self name)
			ifTrue: [ ^symbol ] ].
	self basicCategory: (result := self environment organization categoryOfElement: self name).
	^result! !

!TBehaviorCategorization methodsFor: 'organization' stamp: 'EstebanLorenzano 7/27/2012 16:32'!
category: aString 
	"Categorize the receiver under the system category, aString, removing it from 
	any previous categorization."

	| oldCategory |
	oldCategory := self basicCategory.
	aString isString
		ifTrue: [
			self basicCategory: aString asSymbol.
			self environment organization classify: self name under: self basicCategory ]
		ifFalse: [self errorCategoryName].
	SystemAnnouncer uniqueInstance
		class: self recategorizedFrom: oldCategory to: self basicCategory! !

Trait named: #TClassAndTraitDescription
	uses: {}
	category: 'Traits-Kernel-Traits'!

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'CamilloBruni 8/1/2012 15:57'!
acceptsLoggingOfCompilation
	"Answer whether the receiver's method submisions and class defintions should be logged to the changes file and to the current change set.  The metaclass follows the rule of the class itself
	Weird name is so that it will come lexically before #compile, so that a clean build can make it through."

	^ true! !

!TClassAndTraitDescription methodsFor: 'accessing method dictionary' stamp: 'EstebanLorenzano 8/3/2012 15:14'!
addAndClassifySelector: selector withMethod: compiledMethod inProtocol: category notifying: requestor
	| priorMethodOrNil oldProtocol newProtocol |
	priorMethodOrNil := self compiledMethodAt: selector ifAbsent: [nil].
	self addSelectorSilently: selector withMethod: compiledMethod.
	oldProtocol := self organization categoryOfElement: selector.
	SystemAnnouncer uniqueInstance 
		suspendAllWhile: [self organization classify: selector under: category].
	newProtocol := self organization categoryOfElement: selector.
	priorMethodOrNil isNil
		ifTrue: [SystemAnnouncer uniqueInstance methodAdded: compiledMethod selector: selector inProtocol: category class: self requestor: requestor]
		ifFalse: [SystemAnnouncer uniqueInstance methodChangedFrom: priorMethodOrNil to: compiledMethod selector: selector inClass: self oldProtocol: oldProtocol newProtocol: newProtocol requestor: requestor].! !

!TClassAndTraitDescription methodsFor: 'accessing method dictionary' stamp: 'EstebanLorenzano 7/27/2012 16:33'!
addSelector: selector withMethod: compiledMethod notifying: requestor
	| priorMethodOrNil |
	priorMethodOrNil := self compiledMethodAt: selector ifAbsent: [nil].
	self addSelectorSilently: selector withMethod: compiledMethod.
	priorMethodOrNil isNil
		ifTrue: [SystemAnnouncer uniqueInstance methodAdded: compiledMethod selector: selector inClass: self requestor: requestor]
		ifFalse: [SystemAnnouncer uniqueInstance methodChangedFrom: priorMethodOrNil to: compiledMethod selector: selector inClass: self requestor: requestor].! !

!TClassAndTraitDescription methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 09:51'!
addSelectorSilently: selector withMethod: compiledMethod
	super addSelectorSilently: selector withMethod: compiledMethod.
	self instanceSide noteAddedSelector: selector meta: self isMeta.! !

!TClassAndTraitDescription methodsFor: 'accessing method dictionary' stamp: 'CamilloBruni 1/30/2013 21:59'!
allMethodsInCategory: aName 
	self explicitRequirement.! !

!TClassAndTraitDescription methodsFor: 'organization updating' stamp: 'CamilleTeruel 2/11/2013 14:34'!
applyChangesOfNewTraitCompositionReplacing: oldComposition
	| changedSelectors oldMethodDict |
	oldMethodDict := self methodDict copy.
	changedSelectors := super applyChangesOfNewTraitCompositionReplacing: oldComposition.
	self noteRecategorizedSelectors: changedSelectors oldComposition: oldComposition.
	self noteChangesFrom: oldMethodDict.
	^ changedSelectors.! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 17:13'!
classComment: aString
	"Store the comment, aString or Text or RemoteString, associated with the class we are orgainzing.  Empty string gets stored only if had a non-empty one before."
	^ self classComment: aString stamp: '<historical>'! !

!TClassAndTraitDescription methodsFor: 'fileIn/Out' stamp: 'EstebanLorenzano 7/27/2012 16:33'!
classComment: aString stamp: aStamp
	"Store the comment, aString or Text or RemoteString, associated with the class we are organizing.  Empty string gets stored only if had a non-empty one before."

	| ptr header file oldCommentRemoteStr oldComment oldStamp |
	oldComment := self organization classComment.
	oldStamp := self organization commentStamp.
	(aString isKindOf: RemoteString) ifTrue:
		[SystemAnnouncer uniqueInstance class: self oldComment: oldComment newComment: aString string oldStamp: oldStamp newStamp: aStamp.
		^ self organization classComment: aString stamp: aStamp].

	oldCommentRemoteStr := self organization commentRemoteStr.
	(aString size = 0) & (oldCommentRemoteStr isNil) ifTrue: [^ self organization classComment: nil].
		"never had a class comment, no need to write empty string out"

	ptr := oldCommentRemoteStr ifNil: [0] ifNotNil: [oldCommentRemoteStr sourcePointer].
	SourceFiles ifNotNil: [(file := SourceFiles at: 2) ifNotNil:
		[file setToEnd; cr; nextPut: $!!.	"directly"
		"Should be saying (file command: 'H3') for HTML, but ignoring it here"
		header := String streamContents: [:strm | strm nextPutAll: self name;
			nextPutAll: ' commentStamp: '.
			aStamp storeOn: strm.
			strm nextPutAll: ' prior: '; nextPutAll: ptr printString].
		file nextChunkPut: header]].
	self organization classComment: (RemoteString newString: aString onFileNumber: 2) stamp: aStamp.
	SystemAnnouncer uniqueInstance class: self oldComment: oldComment newComment: aString oldStamp: oldStamp newStamp: aStamp! !

!TClassAndTraitDescription methodsFor: 'accessing comment' stamp: 'dvf 9/27/2005 16:55'!
comment
	"Answer the receiver's comment. (If missing, supply a template) "
	| aString |
	aString := self instanceSide organization classComment.
	aString isEmpty ifFalse: [^ aString].
	^self classCommentBlank! !

!TClassAndTraitDescription methodsFor: 'accessing comment' stamp: 'al 5/9/2004 17:05'!
comment: aStringOrText
	"Set the receiver's comment to be the argument, aStringOrText."

	self instanceSide classComment: aStringOrText.! !

!TClassAndTraitDescription methodsFor: 'accessing comment' stamp: 'al 5/9/2004 17:07'!
comment: aStringOrText stamp: aStamp
	"Set the receiver's comment to be the argument, aStringOrText."

	self instanceSide classComment: aStringOrText stamp: aStamp.! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 11/28/2005 22:18'!
commentStamp: changeStamp
	self organization commentStamp: changeStamp.
	^ self commentStamp: changeStamp prior: 0! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 17:15'!
commentStamp: changeStamp prior: indexAndOffset
	"Prior source link ignored when filing in."

	^ ClassCommentReader new setClass: self
				category: #Comment
				changeStamp: changeStamp! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 11/28/2005 22:18'!
compile: code classified: heading 
	"Compile the argument, code, as source code in the context of the 
	receiver and install the result in the receiver's method dictionary under 
	the classification indicated by the second argument, heading. nil is to be 
	notified if an error occurs. The argument code is either a string or an 
	object that converts to a string or a PositionableStream on an object that 
	converts to a string."

	^self
		compile: code
		classified: heading
		notifying: nil! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'JB 2/12/2010 15:21'!
compile: text classified: category notifying: requestor
	| stamp |
	stamp := self acceptsLoggingOfCompilation ifTrue: [Author changeStamp] ifFalse: [nil].
	^ self compile: text classified: category
		withStamp: stamp notifying: requestor! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 5/8/2004 20:00'!
compile: text classified: category withStamp: changeStamp notifying: requestor
	^ self compile: text classified: category withStamp: changeStamp notifying: requestor logSource: self acceptsLoggingOfCompilation! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'ClementBera 2/8/2013 15:14'!
compile: text classified: category withStamp: changeStamp notifying: requestor logSource: logSource

	| methodAndNode |

	methodAndNode := self
		compile: text asString 
		classified: category
		notifying: requestor
		trailer: self defaultMethodTrailer 
		ifFail: [ ^ nil ].
		
	logSource ifTrue: [
		self 
			logMethodSource: text asString
			forMethodWithNode: methodAndNode
			inCategory: category 
			withStamp: changeStamp 
			notifying: requestor ].
		
	self 
		addAndClassifySelector: methodAndNode selector 
		withMethod: methodAndNode method 
		inProtocol: category 
		notifying: requestor.
			
	self instanceSide 
		noteCompilationOf: methodAndNode selector 
		meta: self isClassSide.

	^ methodAndNode selector! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 5/8/2004 20:12'!
compile: code notifying: requestor 
	"Refer to the comment in Behavior|compile:notifying:." 

	^self compile: code
		 classified: ClassOrganizer default
		 notifying: requestor! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'StephaneDucasse 4/11/2010 23:45'!
compileSilently: code
	"Compile the code and classify the resulting method in the given category, leaving no trail in the system log, nor in any change set, nor in the 'recent submissions' list. This should only be used when you know for sure that the compilation will succeed."

	^ self compileSilently: code classified: '' notifying: nil.! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 5/8/2004 20:12'!
compileSilently: code classified: category
	"Compile the code and classify the resulting method in the given category, leaving no trail in the system log, nor in any change set, nor in the 'recent submissions' list. This should only be used when you know for sure that the compilation will succeed."

	^ self compileSilently: code classified: category notifying: nil.! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'CamilleTeruel 2/19/2013 18:10'!
compileSilently: code classified: category notifying: requestor
	"Compile the code and classify the resulting method in the given category, leaving no trail in the system log, nor in any change set, nor in the 'recent submissions' list. This should only be used when you know for sure that the compilation will succeed."

	^ SystemAnnouncer uniqueInstance 
		suspendAllWhile: [self compile: code classified: category notifying: requestor].! !

!TClassAndTraitDescription methodsFor: 'copying' stamp: 'al 5/8/2004 18:53'!
copy: sel from: class 
	"Install the method associated with the first argument, sel, a message 
	selector, found in the method dictionary of the second argument, class, 
	as one of the receiver's methods. Classify the message under -As yet not 
	classified-."

	self copy: sel
		from: class
		classified: nil! !

!TClassAndTraitDescription methodsFor: 'copying' stamp: 'BenjaminVanRyseghem 3/14/2012 17:35'!
copy: sel from: class classified: cat 
	"Install the method associated with the first arugment, sel, a message 
	selector, found in the method dictionary of the second argument, class, 
	as one of the receiver's methods. Classify the message under the third 
	argument, cat."

	| code category |
	"Useful when modifying an existing class"
	code := class sourceCodeAt: sel.
	code ifNotNil: 
			[cat
				ifNil: [category := class organization categoryOfElement: sel]
				ifNotNil: [category := cat].
			(self includesLocalSelector: sel)
				ifTrue: [code asString = (self sourceCodeAt: sel) asString 
							ifFalse: [self error: self name 
										, ' ' 
										, sel 
										, ' will be redefined if you proceed.']].
			self compile: code classified: category]! !

!TClassAndTraitDescription methodsFor: 'copying' stamp: 'al 5/8/2004 18:54'!
copyAll: selArray from: class 
	"Install all the methods found in the method dictionary of the second 
	argument, class, as the receiver's methods. Classify the messages under 
	-As yet not classified-."

	self copyAll: selArray
		from: class
		classified: nil! !

!TClassAndTraitDescription methodsFor: 'copying' stamp: 'AdrianLienahrd 6/20/2011 11:43'!
copyAll: selArray from: class classified: cat 
	"Install all the methods found in the method dictionary of the second 
	argument, class, as the receiver's methods. Classify the messages under 
	the third argument, cat."

	selArray do: [:s |
		(class includesLocalSelector: s) ifTrue: [
			self copy: s
				from: class
				classified: cat ] ]! !

!TClassAndTraitDescription methodsFor: 'copying' stamp: 'al 5/8/2004 18:54'!
copyAllCategoriesFrom: aClass 
	"Specify that the categories of messages for the receiver include all of 
	those found in the class, aClass. Install each of the messages found in 
	these categories into the method dictionary of the receiver, classified 
	under the appropriate categories."

	aClass organization categories do: [:cat | self copyCategory: cat from: aClass]! !

!TClassAndTraitDescription methodsFor: 'copying' stamp: 'al 5/8/2004 18:54'!
copyCategory: cat from: class 
	"Specify that one of the categories of messages for the receiver is cat, as 
	found in the class, class. Copy each message found in this category."

	self copyCategory: cat
		from: class
		classified: cat! !

!TClassAndTraitDescription methodsFor: 'copying' stamp: 'al 5/8/2004 18:55'!
copyCategory: cat from: aClass classified: newCat 
	"Specify that one of the categories of messages for the receiver is the 
	third argument, newCat. Copy each message found in the category cat in 
	class aClass into this new category."

	self copyAll: (aClass organization listAtCategoryNamed: cat)
		from: aClass
		classified: newCat! !

!TClassAndTraitDescription methodsFor: 'copying' stamp: 'al 5/8/2004 18:55'!
copyMethodDictionaryFrom: donorClass
	"Copy the method dictionary of the donor class over to the receiver"

	self methodDict: donorClass copyOfMethodDictionary.
	self organization: donorClass organization deepCopy.! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'MarcusDenker 2/14/2010 22:20'!
definition
	"Answer a String that defines the receiver"

	^String streamContents: [:stream |
		stream nextPutAll: self class name.
		stream nextPutAll: ' named: ';
				store: self name.
		stream cr; tab; nextPutAll: 'uses: ';
				nextPutAll: self traitCompositionString.
		stream cr; tab; nextPutAll: 'category: ';
				store: self category asString].! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 6/29/2004 21:37'!
doneCompiling
	"A ClassBuilder has finished the compilation of the receiver.
	This message is a notification for a class that needs to do some
	cleanup / reinitialization after it has been recompiled."! !

!TClassAndTraitDescription methodsFor: 'private' stamp: 'al 5/8/2004 19:24'!
errorCategoryName
	self error: 'Category name must be a String'! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'PeterHugossonMiller 9/3/2009 11:32'!
fileOutCategory: catName 

	| internalStream |
	internalStream := (String new: 1000) writeStream.
	internalStream header; timeStamp.
	self fileOutCategory: catName on: internalStream moveSource: false toFile: 0.
	internalStream trailer.

	^ FileStream writeSourceCodeFrom: internalStream baseName: (self name , '-' , catName) isSt: true.! !

!TClassAndTraitDescription methodsFor: 'fileIn/Out' stamp: 'ab 06/29/2011 14:01'!
fileOutCategory: aSymbol on: aFileStream moveSource: moveSource toFile: fileIndex 
	"File a description of the receiver's category, aString, onto aFileStream. If 
	moveSource, is true, then set the method source pointer to the new file position.
	Note when this method is called with moveSource=true, it is condensing the
	.sources file, and should only write one preamble per method category."

	| selectors |
	aFileStream cr.
	selectors := self selectorsToFileOutCategory: aSymbol.

	"Overridden to preserve author stamps in sources file regardless"
	selectors do: [:sel |
		self printMethodChunk: sel 
			withPreamble: true
			on: aFileStream 
			moveSource: moveSource 
			toFile: fileIndex].
	^ self! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:51'!
fileOutChangedMessages: aSet on: aFileStream 
	"File a description of the messages of the receiver that have been 
	changed (i.e., are entered into the argument, aSet) onto aFileStream."

	self fileOutChangedMessages: aSet
		on: aFileStream
		moveSource: false
		toFile: 0! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'nice 1/5/2010 15:59'!
fileOutChangedMessages: aSet on: aFileStream moveSource: moveSource toFile: fileIndex 
	"File a description of the messages of this class that have been 
	changed (i.e., are entered into the argument, aSet) onto aFileStream.  If 
	moveSource, is true, then set the method source pointer to the new file position.
	Note when this method is called with moveSource=true, it is condensing the
	.changes file, and should only write a preamble for every method."
	| org |
	(org := self organization) categories do: 
		[:cat | | sels | 
		sels := (org listAtCategoryNamed: cat) select: [:sel | aSet includes: sel].
		sels do:
			[:sel |  self printMethodChunk: sel withPreamble: true on: aFileStream
							moveSource: moveSource toFile: fileIndex]]! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'PeterHugossonMiller 9/3/2009 11:32'!
fileOutMethod: selector
	"Write source code of a single method on a file.  Make up a name for the file."
	
	| internalStream |
	(selector == #Comment) ifTrue: [^ self inform: 'Sorry, cannot file out class comment in isolation.'].
	(self includesSelector: selector) ifFalse: [^ self error: 'Selector ', selector asString, ' not found'].
	internalStream := (String new: 1000) writeStream.
	internalStream header; timeStamp.
	self printMethodChunk: selector withPreamble: true
		on: internalStream moveSource: false toFile: 0.

	FileStream writeSourceCodeFrom: internalStream baseName: (self name , '-' , (selector copyReplaceAll: ':' with: '')) isSt: true.! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:52'!
fileOutOn: aFileStream 
	"File a description of the receiver on aFileStream."

	self fileOutOn: aFileStream
		moveSource: false
		toFile: 0! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'sd 4/25/2008 15:38'!
fileOutOn: aFileStream moveSource: moveSource toFile: fileIndex
	"File a description of the receiver on aFileStream. If the boolean 
	argument, moveSource, is true, then set the trailing bytes to the position 
	of aFileStream and to fileIndex in order to indicate where to find the 
	source code."

	aFileStream nextChunkPut: self definition.

	self organization
		putCommentOnFile: aFileStream
		numbered: fileIndex
		moveSource: moveSource
		forClass: self.
	self organization categories do: 
		[:heading |
		self fileOutCategory: heading
			on: aFileStream
			moveSource: moveSource
			toFile: fileIndex]! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'BenjaminVanRyseghem 4/12/2012 17:24'!
fileOutOrganizationOn: aFileStream
	"File a description of the receiver's organization on aFileStream."

	aFileStream cr; nextPut: $!!.
	aFileStream nextChunkPut: self name, ' reorganize'; cr.
	aFileStream nextChunkPut: self organization stringForFileOut ; cr! !

!TClassAndTraitDescription methodsFor: 'accessing comment' stamp: 'al 5/9/2004 17:09'!
hasComment
	"return whether this class truly has a comment other than the default"
	| org |
	org := self instanceSide organization.
	^org classComment isEmptyOrNil not! !

!TClassAndTraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'al 5/9/2004 20:50'!
isClassSide
	^self == self classSide! !

!TClassAndTraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'al 5/9/2004 20:50'!
isInstanceSide
	^self isClassSide not! !

!TClassAndTraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'al 5/9/2004 20:55'!
isMeta
	^self isClassSide! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'marcus.denker 8/25/2008 12:04'!
localMethods
	"returns the methods of classes including the ones of the traits that the class uses" 
	 
	^ self methods select: [:each | self includesLocalSelector: each selector].! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'marcus.denker 8/25/2008 12:03'!
methods
	"returns the methods of classes including the ones of the traits that the class uses" 
	 
	^ self methodDict values  ! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:53'!
methodsFor: categoryName 
	"Answer a ClassCategoryReader for compiling the messages in the category, categoryName, of the receiver."

	^ ClassCategoryReader new setClass: self category: categoryName asSymbol

	"(False methodsFor: 'logical operations') inspect"! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:54'!
methodsFor: aString priorSource: sourcePosition inFile: fileIndex
	"Prior source pointer ignored when filing in."
	^ self methodsFor: aString! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'JB 2/12/2010 15:24'!
methodsFor: categoryName stamp: changeStamp 
	^ self methodsFor: categoryName stamp: (Author fixStamp: changeStamp) prior: 0! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:54'!
methodsFor: categoryName stamp: changeStamp prior: indexAndOffset
	"Prior source link ignored when filing in."
	^ ClassCategoryReader new setClass: self
				category: categoryName asSymbol
				changeStamp: changeStamp

"Most importantly, return the new ClassCategoryReader, so a fileIn will let it seize control.  So method will be placed in the proper category.  See the transfer of control where ReadWriteStream fileIn calls scanFrom:"! !

!TClassAndTraitDescription methodsFor: 'accessing method dictionary' stamp: 'CamilloBruni 1/30/2013 22:00'!
methodsInCategory: aName 
	"Answer a list of the methods of the receiver that are in category named aName"
	
	| aColl |
	aColl := Set withAll: (aName = ClassOrganizer allCategory
			ifTrue: [self organization allMethodSelectors]
			ifFalse: [self organization listAtCategoryNamed: aName]).
	^ aColl asArray sort! !

!TClassAndTraitDescription methodsFor: 'fileIn/Out' stamp: 'MarcusDenker 2/21/2010 12:52'!
moveChangesTo: newFile 
	"Used in the process of condensing changes, this message requests that 
	the source code of all methods of the receiver that have been changed 
	should be moved to newFile."

	| changes |
	changes := self selectors select: [:sel | (self compiledMethodAt: sel) fileIndex > 1 ].
	self
		fileOutChangedMessages: changes
		on: newFile
		moveSource: true
		toFile: 2! !

!TClassAndTraitDescription methodsFor: 'accessing method dictionary' stamp: 'al 5/9/2004 16:36'!
noteAddedSelector: aSelector meta: isMeta
	"A hook allowing some classes to react to adding of certain selectors"! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'dvf 9/27/2005 17:13'!
noteCompilationOf: aSelector meta: isMeta
	"A hook allowing some classes to react to recompilation of certain selectors"! !

!TClassAndTraitDescription methodsFor: 'organization updating' stamp: 'al 5/9/2004 14:00'!
noteRecategorizedSelector: aSymbol from: oldCategoryOrNil to: newCategoryOrNil
	| changedCategories |
	changedCategories := self updateOrganizationSelector: aSymbol oldCategory: oldCategoryOrNil newCategory: newCategoryOrNil.
	changedCategories do: [:each |
		(self organization isEmptyCategoryNamed: each) ifTrue: [self organization removeCategory: each]]! !

!TClassAndTraitDescription methodsFor: 'organization updating' stamp: 'nice 1/5/2010 15:59'!
noteRecategorizedSelectors: aCollection oldComposition: aTraitComposition
	
	aCollection do: [:each | | oldCategory newCategory | 
		oldCategory := self organization categoryOfElement: each.
		newCategory := (self traitComposition methodDescriptionForSelector: each) effectiveMethodCategory.
		self noteRecategorizedSelector: each from: oldCategory to: newCategory]! !

!TClassAndTraitDescription methodsFor: 'organization updating' stamp: 'EstebanLorenzano 7/27/2012 16:33'!
notifyOfRecategorizedSelector: element from: oldCategory to: newCategory
	SystemAnnouncer uniqueInstance selector: element recategorizedFrom: oldCategory to: newCategory inClass: self! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:54'!
printCategoryChunk: categoryName on: aFileStream
	^ self printCategoryChunk: categoryName withStamp: '' on: aFileStream! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'JB 2/12/2010 15:21'!
printCategoryChunk: category on: aFileStream priorMethod: priorMethod
	^ self printCategoryChunk: category on: aFileStream
		withStamp: Author changeStamp priorMethod: priorMethod! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'sd 4/25/2008 15:39'!
printCategoryChunk: category on: aFileStream withStamp: changeStamp priorMethod: priorMethod 
	"Print a method category preamble.  This must have a category name.
	It may have an author/date stamp, and it may have a prior source link.
	If it has a prior source link, it MUST have a stamp, even if it is empty."

"The current design is that changeStamps and prior source links are preserved in the changes file.  All fileOuts include changeStamps.  Condensing sources, however, eliminates all stamps (and links, natch)."

	aFileStream cr; nextPut: $!!.
	aFileStream nextChunkPut: (String streamContents:
		[:strm |
		strm nextPutAll: self name; nextPutAll: ' methodsFor: '; print: category asString.
		(changeStamp ~~ nil and:
			[changeStamp size > 0 or: [priorMethod ~~ nil]]) ifTrue:
			[strm nextPutAll: ' stamp: '; print: changeStamp].
		priorMethod ~~ nil ifTrue:
			[strm nextPutAll: ' prior: '; print: priorMethod sourcePointer]]).
	! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:55'!
printCategoryChunk: categoryName withStamp: changeStamp on: aFileStream
	^ self printCategoryChunk: categoryName on: aFileStream withStamp: changeStamp
		priorMethod: nil! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'md 4/26/2012 12:04'!
printMethodChunk: selector withPreamble: doPreamble on: outStream
		moveSource: moveSource toFile: fileIndex
	"Copy the source code for the method associated with selector onto the fileStream.  If moveSource true, then also set the source code pointer of the method."
	| preamble method oldPos newPos sourceFile endPos |
	doPreamble 
		ifTrue: [preamble := self name , ' methodsFor: ' ,
					(self organization categoryOfElement: selector) asString printString]
		ifFalse: [preamble := ''].
	method := self methodDict at: selector ifAbsent:
		[outStream nextPutAll: selector; cr.
		outStream tab; nextPutAll: '** ERROR!!  THIS SCRIPT IS MISSING ** ' translated; cr; cr.
		outStream nextPutAll: '  '.
		^ outStream].

	((method fileIndex = 0
		or: [(SourceFiles at: method fileIndex) == nil])
		or: [(oldPos := method filePosition) = 0])
		ifTrue:
		["The source code is not accessible.  We must decompile..."
		preamble size > 0 ifTrue: [outStream cr; nextPut: $!!; nextChunkPut: preamble; cr].
		outStream nextChunkPut: method decompileString]
		ifFalse:
		[sourceFile := SourceFiles at: method fileIndex.
		preamble size > 0
			ifTrue:    "Copy the preamble"
				[outStream copyPreamble: preamble from: sourceFile at: oldPos]
			ifFalse:
				[sourceFile position: oldPos].
		"Copy the method chunk"
		newPos := outStream position.
		outStream copyMethodChunkFrom: sourceFile.
		sourceFile skipSeparators.      "The following chunk may have ]style["
		sourceFile peek == $] ifTrue: [
			outStream cr; copyMethodChunkFrom: sourceFile].
		moveSource ifTrue:    "Set the new method source pointer"
			[endPos := outStream position.
			method setSourcePosition: newPos inFile: fileIndex]].
	preamble size > 0 ifTrue: [outStream nextChunkPut: ' '].
	^ outStream cr! !

!TClassAndTraitDescription methodsFor: 'printing' stamp: 'al 5/9/2004 16:50'!
printOn: aStream 
	aStream nextPutAll: self name! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'StephaneDucasse 5/28/2011 13:42'!
putClassCommentToCondensedChangesFile: aFileStream
	"Called when condensing changes.  If the receiver has a class comment, and if that class comment does not reside in the .sources file, then write it to the given filestream, with the resulting RemoteString being reachable from the source file #2.  Note that any existing backpointer into the .sources file is lost by this process -- a situation that maybe should be fixed someday."

	| header aStamp aCommentRemoteStr |
	self isMeta ifTrue: [^ self].  "bulletproofing only"
	((aCommentRemoteStr := self organization commentRemoteStr) isNil or:
		[aCommentRemoteStr sourceFileNumber = 1]) ifTrue: [^ self].

	aFileStream cr; nextPut: $!!.
	header := String streamContents: [:strm | strm nextPutAll: self name;
		nextPutAll: ' commentStamp: '.
		(aStamp := self organization commentStamp ifNil: ['<historical>']) storeOn: strm.
		strm nextPutAll: ' prior: 0'].
	aFileStream nextChunkPut: header.
	aFileStream cr.
	self organization classComment: (RemoteString newString: self organization classComment onFileNumber: 2 toFile: aFileStream) stamp: aStamp! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 5/9/2004 19:17'!
reformatAll
	"Reformat all methods in this class.
	Leaves old code accessible to version browsing"
	self selectorsDo: [:sel | self reformatMethodAt: sel]! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'alain.plantec 5/18/2009 16:01'!
reformatMethodAt: selector
	| newCodeString method |
	newCodeString := self prettyPrinterClass 
				format: (self sourceCodeAt: selector)
				in: self
				notifying: nil.
	method := self compiledMethodAt: selector.
	method 
		putSource: newCodeString
		fromParseNode: nil
		class: self
		category: (self organization categoryOfElement: selector)
		inFile: 2
		priorMethod: method
! !

!TClassAndTraitDescription methodsFor: 'accessing method dictionary' stamp: 'al 5/8/2004 19:50'!
removeCategory: aString 
	"Remove each of the messages categorized under aString in the method 
	dictionary of the receiver. Then remove the category aString."
	| categoryName |
	categoryName := aString asSymbol.
	(self organization listAtCategoryNamed: categoryName) do:
		[:sel | self removeSelector: sel].
	self organization removeCategory: categoryName! !

!TClassAndTraitDescription methodsFor: 'accessing method dictionary' stamp: 'MartinDias 2/11/2013 14:08'!
removeSelector: selector 
	"Remove the message whose selector is given from the method 
	dictionary of the receiver, if it is there. Answer nil otherwise."
	
	| priorMethod priorProtocol origin | 
	priorMethod := self compiledMethodAt: selector ifAbsent: [^ nil].
	origin := priorMethod origin.
	priorProtocol := self whichCategoryIncludesSelector: selector.
	super removeSelector: selector.
	SystemAnnouncer uniqueInstance 
		suspendAllWhile: [self updateOrganizationSelector: selector oldCategory: priorProtocol newCategory: nil].
	SystemAnnouncer uniqueInstance 
			methodRemoved: priorMethod selector: selector inProtocol: priorProtocol class: self origin: origin.! !

!TClassAndTraitDescription methodsFor: 'organization' stamp: 'al 5/9/2004 19:19'!
reorganize
	"During fileIn, !!Rectangle reorganize!! allows Rectangle to seize control and treat the next chunk as its organization.  See the transfer of control where ReadWriteStream fileIn calls scanFrom:"

	^self organization! !

!TClassAndTraitDescription methodsFor: 'fileIn/Out' stamp: 'ab 06/29/2011 14:01'!
selectorsToFileOutCategory: aSymbol
	^ ((aSymbol asString = ClassOrganizer allCategory)
				ifTrue: [ self organization allMethodSelectors ]
				ifFalse: [ self organization listAtCategoryNamed: aSymbol ]) select: [:sel |
					self includesLocalSelector: sel ]! !

!TClassAndTraitDescription methodsFor: 'printing' stamp: 'al 5/9/2004 16:50'!
storeOn: aStream
	"Classes and Metaclasses have global names."

	aStream nextPutAll: self name! !

!TClassAndTraitDescription methodsFor: 'accessing method dictionary' stamp: 'CamilloBruni 1/30/2013 22:04'!
uncategorizedMethods
	^ self methodsInCategory: ClassOrganizer default! !

!TClassAndTraitDescription methodsFor: 'organization updating' stamp: 'nice 1/5/2010 15:59'!
updateOrganizationSelector: aSymbol oldCategory: oldCategoryOrNil newCategory: newCategoryOrNil
	| changedCategories composition |
	changedCategories := IdentitySet new.
	composition := self hasTraitComposition
		ifTrue: [self traitComposition]
		ifFalse: [TraitComposition new].
	(composition methodDescriptionsForSelector: aSymbol) do: [:each | | effectiveCategory sel currentCategory |
		sel := each selector.
		(self includesLocalSelector: sel) ifFalse: [
			currentCategory := self organization categoryOfElement: sel.
			effectiveCategory := each effectiveMethodCategoryCurrent: currentCategory new: newCategoryOrNil.
			effectiveCategory isNil ifTrue: [
				currentCategory ifNotNil: [changedCategories add: currentCategory].
				self organization removeElement: sel.
			] ifFalse: [
				((currentCategory isNil or: [currentCategory == ClassOrganizer ambiguous or: [currentCategory == oldCategoryOrNil]]) and: [currentCategory ~~ effectiveCategory]) ifTrue: [
					currentCategory ifNotNil: [changedCategories add: currentCategory].
					self organization 
						classify: sel 
						under: effectiveCategory
						suppressIfDefault: false]]]].
	^ changedCategories! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'CamilloBruni 8/1/2012 16:18'!
wantsChangeSetLogging
	"Answer whether code submitted for the receiver should be remembered by the changeSet mechanism."

	^ true! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 5/8/2004 20:13'!
wantsRecompilationProgressReported
	"Answer whether the receiver would like progress of its recompilation reported interactively to the user."

	^ true! !

!TClassAndTraitDescription methodsFor: 'organization' stamp: 'al 5/8/2004 19:15'!
whichCategoryIncludesSelector: aSelector 
	"Answer the category of the argument, aSelector, in the organization of 
	the receiver, or answer nil if the receiver does not inlcude this selector."

	(self includesSelector: aSelector)
		ifTrue: [^ self organization categoryOfElement: aSelector]
		ifFalse: [^nil]! !

!TClassAndTraitDescription methodsFor: 'organization' stamp: 'al 5/8/2004 19:29'!
zapOrganization
	"Remove the organization of this class by message categories.
	This is typically done to save space in small systems.  Classes and methods
	created or filed in subsequently will, nonetheless, be organized"

	self organization: nil.
	self isClassSide ifFalse: [self classSide zapOrganization]! !

Trait named: #TComposingDescription
	uses: {}
	category: 'Traits-Kernel-Traits'!

!TComposingDescription methodsFor: 'composition' stamp: 'apb 8/22/2005 17:31'!
+ aTraitOrTraitComposition
	"Use double dispatch to avoid having nested composition in cases where
	parenthesis are used, such as T1 + (T2 + T3)"
	
	^aTraitOrTraitComposition addOnTheLeft: self! !

!TComposingDescription methodsFor: 'composition' stamp: 'apb 8/22/2005 17:12'!
- anArrayOfSelectors
	^TraitExclusion
		with: self
		exclusions: anArrayOfSelectors! !

!TComposingDescription methodsFor: 'composition' stamp: 'apb 8/22/2005 17:02'!
@ anArrayOfAssociations 
	^ TraitAlias with: self aliases: anArrayOfAssociations! !

!TComposingDescription methodsFor: 'private' stamp: 'apb 8/22/2005 17:53'!
addCompositionOnLeft: aTraitComposition
	^ aTraitComposition add: self! !

!TComposingDescription methodsFor: 'private' stamp: 'apb 8/22/2005 17:40'!
addOnTheLeft: aTraitExpression
	^TraitComposition with: aTraitExpression with: self! !

!TComposingDescription methodsFor: 'converting' stamp: 'apb 8/22/2005 15:47'!
asTraitComposition
	^TraitComposition with: self! !

Trait named: #TPureBehavior
	uses: {}
	category: 'Traits-Kernel-Traits'!

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 8/2/2004 16:26'!
>> selector 
	"Answer the compiled method associated with the argument, selector (a 
	Symbol), a message selector in the receiver's method dictionary. If the 
	selector is not in the dictionary, create an error notification."

	^self compiledMethodAt: selector
! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 7/30/2004 21:51'!
addExclusionOf: aSymbol to: aTrait
	self setTraitComposition: (
		self traitComposition copyWithExclusionOf: aSymbol to: aTrait)! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'al 7/30/2004 22:43'!
addSelector: selector withMethod: compiledMethod 
	^ self addSelector: selector withMethod: compiledMethod notifying: nil! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'al 7/30/2004 22:43'!
addSelector: selector withMethod: compiledMethod notifying: requestor
	^ self addSelectorSilently: selector withMethod: compiledMethod! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'al 9/16/2005 14:31'!
addSelectorSilently: selector withMethod: compiledMethod
	self methodDictAddSelectorSilently: selector withMethod: compiledMethod.
	self registerLocalSelector: selector! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 1/9/2006 17:58'!
addToComposition: aTrait
	self setTraitComposition: (self traitComposition copyTraitExpression
		add: aTrait;
		yourself)! !

!TPureBehavior methodsFor: 'traits' stamp: 'CamilloBruni 4/27/2012 18:22'!
addTraitSelector: aSymbol withMethod: aCompiledMethod
	"Add aMethod with selector aSymbol to my
	methodDict. aMethod must not be defined locally."

	| source methodAndNode |
	[(self includesLocalSelector: aSymbol) not] assert.
	self ensureLocalSelectors.
		
	source := aCompiledMethod getSourceReplacingSelectorWith: aSymbol.
	methodAndNode  := self
		compile: source
		classified: aCompiledMethod category
		notifying: nil
		trailer: self defaultMethodTrailer
		ifFail: [^nil].
	methodAndNode method putSource: source fromParseNode: methodAndNode node inFile: 2
		withPreamble: [:f | f cr; nextPut: $!!; nextChunkPut: 'Trait method'; cr].
			
	self basicAddSelector: aSymbol withMethod: methodAndNode method! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:12'!
allSelectors
	self explicitRequirement! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 7/30/2004 21:39'!
applyChangesOfNewTraitCompositionReplacing: oldComposition
	| changedSelectors |
	changedSelectors := self traitComposition
		changedSelectorsComparedTo: oldComposition.
	changedSelectors isEmpty ifFalse: [
		self noteChangedSelectors: changedSelectors].
	self traitComposition isEmpty ifTrue: [
		self purgeLocalSelectors].
	^changedSelectors! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'MarcusDenker 2/26/2012 12:52'!
basicAddSelector: selector withMethod: compiledMethod 
	"Add the message selector with the corresponding compiled method to the 
	receiver's method dictionary.
	Do this without sending system change notifications"

	| oldMethodOrNil |	
	oldMethodOrNil := self lookupSelector: selector.
	self methodDict at: selector put: compiledMethod.
	compiledMethod methodClass: self.
	compiledMethod selector: selector.

	"Now flush the method cache, when we add a method"
	oldMethodOrNil ifNotNil: [oldMethodOrNil flushCache].! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:13'!
basicLocalSelectors
	self explicitRequirement! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:13'!
basicLocalSelectors: aSetOrNil
	self explicitRequirement! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'MarcusDenker 2/26/2012 12:53'!
basicRemoveSelector: selector 
	"Assuming that the argument, selector (a Symbol), is a message selector 
	in my method dictionary, remove it and its method."

	| oldMethod |
	oldMethod := self methodDict at: selector ifAbsent: [^ self].
	self methodDict removeKey: selector.

	"Now flush the method cache"
	oldMethod flushCache.! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'al 9/16/2005 14:16'!
canUnderstand: selector
	"Answer whether the receiver can respond to the message whose selector 
	is the argument."

	^ self classAndMethodFor: selector do: [:c :m | m isProvided] ifAbsent: [false].! !

!TPureBehavior methodsFor: 'testing' stamp: 'al 9/16/2005 14:23'!
canZapMethodDictionary
	"Return true if it is safe to zap the method dictionary on #obsolete"
	^true! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 11/28/2005 21:48'!
changeRecordsAt: selector
	"Return a list of ChangeRecords for all versions of the method at selector. Source code can be retrieved by sending string to any one.  Return nil if the method is absent."

	"(Pen changeRecordsAt: #go:) collect: [:cRec | cRec string]"
	^ChangeSet 
		scanVersionsOf: (self compiledMethodAt: selector ifAbsent: [^ nil])
		class: self meta: self isMeta
		category: (self whichCategoryIncludesSelector: selector)
		selector: selector.! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:14'!
classAndMethodFor: aSymbol do: binaryBlock ifAbsent: absentBlock
	"Looks up the selector aSymbol in this class/trait. If it is found, binaryBlock is evaluated
	with this class/trait and the associated method. Otherwise absentBlock is evaluated.
	Note that this implementation is very simple because PureBehavior does not know
	about inheritance (cf. implementation in Behavior)"
	
	^ binaryBlock value: self value: (self compiledMethodAt: aSymbol ifAbsent: [^ absentBlock value]).! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 8/2/2004 16:29'!
compiledMethodAt: selector 
	"Answer the compiled method associated with the argument, selector (a 
	Symbol), a message selector in the receiver's method dictionary. If the 
	selector is not in the dictionary, create an error notification."

	^ self methodDict at: selector! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 20:55'!
compiledMethodAt: selector ifAbsent: aBlock
	"Answer the compiled method associated with the argument, selector (a Symbol), a message selector in the receiver's method dictionary. If the selector is not in the dictionary, return the value of aBlock"

	^ self methodDict at: selector ifAbsent: [aBlock value]! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 20:55'!
compress
	"Compact the method dictionary of the receiver."

	self methodDict rehash! !

!TPureBehavior methodsFor: 'copying' stamp: 'al 7/30/2004 22:51'!
copyOfMethodDictionary
	"Return a copy of the receiver's method dictionary"

	^ self methodDict copy! !

!TPureBehavior methodsFor: 'copying' stamp: 'al 9/16/2005 14:20'!
deepCopy
	"Classes should only be shallowCopied or made anew."

	^ self shallowCopy! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 11/28/2005 11:46'!
deregisterLocalSelector: aSymbol
	self basicLocalSelectors notNil ifTrue: [
		self basicLocalSelectors remove: aSymbol ifAbsent: []]! !

!TPureBehavior methodsFor: 'initialization' stamp: 'al 12/1/2005 16:01'!
emptyMethodDictionary

	^ MethodDictionary new! !

!TPureBehavior methodsFor: 'traits' stamp: 'nice 12/4/2009 00:06'!
ensureLocalSelectors
	"Ensures that the instance variable localSelectors is effectively used to maintain
	the set of local selectors.
	This method must be called before any non-local selectors are added to the
	method dictionary!!"

	self basicLocalSelectors isNil 
		ifTrue: [self basicLocalSelectors: self selectors asSet]! !

!TPureBehavior methodsFor: 'naming' stamp: 'StephaneDucasse 3/20/2010 17:24'!
environment
	"Return the environment in which the receiver is visible"
	^Smalltalk globals! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 12/1/2005 15:40'!
firstCommentAt:  selector
	"Answer a string representing the first comment in the method associated with selector.  Return an empty string if the relevant source file is not available, or if the method's source code does not contain a comment.  Not smart enough to bypass quotes in string constants, but does map doubled quote into a single quote."

	|someComments|
	someComments := self commentsAt: selector.
	^someComments isEmpty ifTrue: [''] ifFalse: [someComments first]


"Behavior firstCommentAt: #firstCommentAt:"! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'PeterHugossonMiller 9/2/2009 16:00'!
firstPrecodeCommentFor:  selector
	"If there is a comment in the source code at the given selector that preceeds the body of the method, return it here, else return nil"

	| parser source tree |
	"Behavior firstPrecodeCommentFor: #firstPrecodeCommentFor:"
	(#(Comment Definition Hierarchy) includes: selector)
		ifTrue:
			["Not really a selector"
			^ nil].
	source := self sourceCodeAt: selector asSymbol ifAbsent: [^ nil].
	parser := self parserClass new.
	tree := 
		parser
			parse: source readStream
			class: self
			noPattern: false
			context: nil
			notifying: nil
			ifFail: [^ nil].
	^ (tree comment ifNil: [^ nil]) first! !

!TPureBehavior methodsFor: 'traits' stamp: 'jannik.laval 5/1/2010 16:32'!
flattenDown: aTrait
	| selectors |
	[self hasTraitComposition and: [self traitComposition allTraits includes: aTrait]] assert.
	selectors := (self traitComposition transformationOfTrait: aTrait) selectors.
	self basicLocalSelectors: self basicLocalSelectors , selectors.
	self removeFromComposition: aTrait.! !

!TPureBehavior methodsFor: 'traits' stamp: 'jannik.laval 5/1/2010 16:32'!
flattenDownAllTraits
	self traitComposition allTraits do: [:each | self flattenDown: each].
	[ self traitComposition isEmpty ] assert.
	self traitComposition: nil.! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:49'!
"popeye" formalHeaderPartsFor: "olive oil" aSelector
	"RELAX!!  The warning you may have just seen about possibly having a bad source file does not apply here, because this method *intends* to have its source code start with a comment.
	This method returns a collection giving the parts in the formal declaration for aSelector.  This parse is in support of schemes in which adjutant properties of a method can be declared via special comments secreted in the formal header
	The result will have
     	3 elements for a simple, argumentless selector.
		5 elements for a single-argument selector
		9 elements for a two-argument selector
		13 elements for a three-argument, selector
		etc...

	The syntactic elements are:

		1		comment preceding initial selector fragment

		2		first selector fragment
		3		comment following first selector fragment  (nil if selector has no arguments)

        ----------------------  (ends here for, e.g., #copy)

		4		first formal argument
		5		comment following first formal argument (nil if selector has only one argument)

        ----------------------  (ends here for, e.g., #copyFrom:)

		6		second keyword
		7		comment following second keyword
		8		second formal argument
		9		comment following second formal argument (nil if selector has only two arguments)

         ----------------------  (ends here for, e.g., #copyFrom:to:)

	Any nil element signifies an absent comment.
	NOTE: The comment following the final formal argument is *not* successfully retrieved by this method in its current form, though it can be obtained, if needed, by other means (e.g. calling #firstPrecodeCommentFor:).  Thus, the *final* element in the structure returned by this method is always going to be nil."

	^ Scanner new scanMessageParts: (self methodHeaderFor: aSelector)

"
	Behavior class formalHeaderPartsFor: #formalHeaderPartsFor:
"


	! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'nice 12/25/2009 20:54'!
formalParametersAt: aSelector
	"Return the names of the arguments used in this method."

	| source |
	source := self sourceCodeAt: aSelector ifAbsent: [^ #()].	"for now"
	^(self parserClass new) parseParameterNames: source! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'marcus.denker 8/25/2008 09:23'!
hasMethods
	"Answer whether the receiver has any methods in its method dictionary."

	^ self methodDict notEmpty! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 9/16/2005 14:29'!
hasTraitComposition
	self explicitRequirement! !

!TPureBehavior methodsFor: 'testing' stamp: 'al 3/1/2006 23:02'!
includesBehavior: aBehavior
	^self == aBehavior! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'al 9/16/2005 14:16'!
includesLocalSelector: aSymbol
	^self basicLocalSelectors isNil
		ifTrue: [self includesSelector: aSymbol]
		ifFalse: [self localSelectors includes: aSymbol]! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'al 7/30/2004 22:35'!
includesSelector: aSymbol 
	"Answer whether the message whose selector is the argument is in the 
	method dictionary of the receiver's class."

	^ self methodDict includesKey: aSymbol! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'StephaneDucasse 4/27/2010 11:54'!
isAliasSelector: aSymbol
	"Return true if the selector aSymbol is an alias defined
	in my or in another composition somewhere deeper in 
	the tree of traits compositions."

	^(self includesLocalSelector: aSymbol) not
		and: [self hasTraitComposition
				and: [self traitComposition isAliasSelector: aSymbol]]! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'al 9/16/2005 14:16'!
isDisabledSelector: selector
	^ self classAndMethodFor: selector do: [:c :m | m isDisabled] ifAbsent: [false]! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'StephaneDucasse 4/27/2010 11:55'!
isLocalAliasSelector: aSymbol
	"Return true if the selector aSymbol is an alias defined
	in my trait composition."

	^(self includesLocalSelector: aSymbol) not
		and: [self hasTraitComposition
				and: [self traitComposition isLocalAliasSelector: aSymbol]]! !

!TPureBehavior methodsFor: 'printing' stamp: 'marcus.denker 11/10/2008 10:04'!
literalScannedAs: scannedLiteral notifying: requestor
	"Postprocesses a literal scanned by Scanner scanToken (esp. xLitQuote).
	If scannedLiteral is not an association, answer it.
	Else, if it is of the form:
		nil->#NameOfMetaclass
	answer nil->theMetaclass, if any has that name, else report an error.
	Else, if it is of the form:
		#NameOfGlobalVariable->anythiEng
	answer the global, class, or pool association with that nameE, if any, else
	add it to Undeclared a answer the new Association."

	| key value |
	(scannedLiteral isVariableBinding)
		ifFalse: [^ scannedLiteral].
	key := scannedLiteral key.
	value := scannedLiteral value.
	key isNil 
		ifTrue: "###<metaclass soleInstance name>"
			[(self bindingOf: value) ifNotNil:[:assoc|
				 (assoc value isKindOf: Behavior)
					ifTrue: [^ nil->assoc value class]].
			 requestor notify: 'No such metaclass'.
			 ^false].
	(key isSymbol)
		ifTrue: "##<global var name>"
			[(self bindingOf: key) ifNotNil:[:assoc | ^assoc].
			Undeclared at: key put: nil.
			 ^Undeclared bindingOf: key].
	requestor notify: '## must be followed by a non-local variable name'.
	^false

"	Form literalScannedAs: 14 notifying: nil 14
	Form literalScannedAs: #OneBitForm notiEfying: nil  OneBitForm
	Form literalScannedAs: ##OneBitForm notifying: nil  OneBitForm->a Form
	Form literalScannedAs: ##Form notifying: nil   Form->Form
	Form literalScannedAs: ###Form notifying: nil   nilE->Form class
"! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'nice 12/4/2009 00:07'!
localSelectors
	"Return a set of selectors defined locally.
	The instance variable is lazily initialized. If it is nil then there
	are no non-local selectors"

	^ self basicLocalSelectors isNil
		ifTrue: [self selectors asSet]
		ifFalse: [self basicLocalSelectors].! !

!TPureBehavior methodsFor: 'printing' stamp: 'al 9/16/2005 14:21'!
longPrintOn: aStream
	"Append to the argument, aStream, the names and values of all of the receiver's instance variables.  But, not useful for a class with a method dictionary."

	aStream nextPutAll: '<<too complex to show>>'; cr.! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:50'!
lookupSelector: selector
	^ self explicitRequirement! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:50'!
methodDict
	^ self explicitRequirement! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:50'!
methodDict: aDictionary
	^ self explicitRequirement! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'adrian_lienhard 2/17/2009 22:00'!
methodDictAddSelectorSilently: selector withMethod: compiledMethod
	self basicAddSelector: selector withMethod: compiledMethod! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 21:02'!
methodDictionary
	"Convenience"
	^self methodDict! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/30/2005 18:24'!
methodDictionary: aDictionary
	self methodDict: aDictionary! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:50'!
methodHeaderFor: selector 
	"Answer the string corresponding to the method header for the given selector"

	| sourceString parser |
	sourceString := self ultimateSourceCodeAt: selector ifAbsent: [self standardMethodHeaderFor: selector].
	(parser := self parserClass new) parseSelector: sourceString.
	^ sourceString asString copyFrom: 1 to: (parser endOfLastToken min: sourceString size)

	"Behavior methodHeaderFor: #methodHeaderFor: "
! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 21:02'!
methodsDo: aBlock
	"Evaluate aBlock for all the compiled methods in my method dictionary."

	^ self methodDict valuesDo: aBlock! !

!TPureBehavior methodsFor: 'naming' stamp: 'al 9/16/2005 14:21'!
name
	^ self explicitRequirement! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 8/6/2004 13:09'!
noteChangedSelectors: aCollection
	"Start update of my methodDict (after changes to traits in traitComposition
	or after a local method was removed from my methodDict). The argument 
	is a collection of method selectors that may have been changed. Most of the time
	aCollection only holds one selector. But when there are aliases involved 
	there may be several method changes that have to be propagated to users."

	| affectedSelectors |
	affectedSelectors := IdentitySet new.
	aCollection do: [:selector |
		affectedSelectors addAll: (self updateMethodDictionarySelector: selector)].
	self notifyUsersOfChangedSelectors: affectedSelectors.
	^ affectedSelectors! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 7/30/2004 21:40'!
notifyUsersOfChangedSelector: aSelector
	self notifyUsersOfChangedSelectors: (Array with: aSelector)! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 7/30/2004 21:40'!
notifyUsersOfChangedSelectors: aCollection! !

!TPureBehavior methodsFor: 'initialization' stamp: 'al 12/1/2005 15:24'!
obsolete
	"Invalidate and recycle local methods,
	e.g., zap the method dictionary if can be done safely."
	self canZapMethodDictionary
		ifTrue: [self methodDict: self emptyMethodDictionary].
	self hasTraitComposition ifTrue: [
		self traitComposition traits do: [:each |
			each removeUser: self]]! !

!TPureBehavior methodsFor: 'copying' stamp: 'nice 1/13/2010 21:59'!
postCopy
	super postCopy.
	self methodDictionary: self copyOfMethodDictionary! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 11/28/2005 11:47'!
purgeLocalSelectors
	self basicLocalSelectors: nil! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 11/28/2005 11:47'!
registerLocalSelector: aSymbol
	self basicLocalSelectors notNil ifTrue: [
		self basicLocalSelectors add: aSymbol]! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 7/30/2004 21:54'!
removeAlias: aSymbol of: aTrait
	self setTraitComposition: (
		self traitComposition copyWithoutAlias: aSymbol of: aTrait)! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 10/22/2006 18:22'!
removeFromComposition: aTrait
	self setTraitComposition: (self traitComposition copyTraitExpression
		removeFromComposition: aTrait)! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'StephaneDucasse 4/27/2010 11:55'!
removeSelector: aSelector 
	"Assuming that the argument, selector (a Symbol), is a message selector 
	in my method dictionary, remove it and its method.
	
	If the method to remove will be replaced by a method from my trait composition,
	the current method does not have to be removed because we mark it as non-local.
	If it is not identical to the actual method from the trait it will be replaced automatically
	by #noteChangedSelectors:.
	
	This is useful to avoid bootstrapping problems when moving methods to a trait
	(e.g., from TPureBehavior to TMethodDictionaryBehavior). Manual moving (implementing
	the method in the trait and then remove it from the class) does not work if the methods
	themselves are used for this process (such as compiledMethodAt:, includesLocalSelector: or
	addTraitSelector:withMethod:)"

	| changeFromLocalToTraitMethod |
	changeFromLocalToTraitMethod := (self includesLocalSelector: aSelector)
		and: [self hasTraitComposition
				and: [self traitComposition includesMethod: aSelector]].

	changeFromLocalToTraitMethod
		ifFalse: [self basicRemoveSelector: aSelector]
		ifTrue: [self ensureLocalSelectors].
	self deregisterLocalSelector: aSelector.
	self noteChangedSelectors: (Array with: aSelector)
	
! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'EstebanLorenzano 8/3/2012 15:29'!
removeSelectorSilently: selector 
	"Remove selector without sending system change notifications"

	^ SystemAnnouncer uniqueInstance suspendAllWhile: [self removeSelector: selector].! !

!TPureBehavior methodsFor: 'traits' stamp: 'jannik.laval 5/1/2010 16:32'!
removeTraitSelector: aSymbol
	[(self includesLocalSelector: aSymbol) not] assert.
	self basicRemoveSelector: aSymbol! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 21:03'!
selectors
	"Answer a Set of all the message selectors specified in the receiver's 
	method dictionary."

	^ self methodDict keys! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'nice 1/16/2010 12:50'!
selectorsAndMethodsDo: selectorAndMethodBlock
	"Evaluate selectorAndMethodBlock with two arguments for each selector/method pair in my method dictionary."

	^ self methodDict keysAndValuesDo: selectorAndMethodBlock! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 21:03'!
selectorsDo: selectorBlock
	"Evaluate selectorBlock for all the message selectors in my method dictionary."

	^ self methodDict keysDo: selectorBlock! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'nice 1/5/2010 15:59'!
selectorsWithArgs: numberOfArgs
	"Return all selectors defined in this class that take this number of arguments.  Could use String.keywords.  Could see how compiler does this."

	| list |
	list := OrderedCollection new.
	self selectorsDo: [:aSel | | num | 
		num := aSel count: [:char | char == $:].
		num = 0 ifTrue: [aSel last isLetter ifFalse: [num := 1]].
		num = numberOfArgs ifTrue: [list add: aSel]].
	^ list! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 7/30/2004 21:48'!
setTraitComposition: aTraitComposition
	| oldComposition |
	(self hasTraitComposition not and: [aTraitComposition isEmpty]) ifTrue: [^self].
	aTraitComposition assertValidUser: self.

	oldComposition := self traitComposition.
	self traitComposition: aTraitComposition.
	self applyChangesOfNewTraitCompositionReplacing: oldComposition.
	
	oldComposition traits do: [:each | each removeUser: self].
	aTraitComposition traits do: [:each | each addUser: self]! !

!TPureBehavior methodsFor: 'traits' stamp: 'apb 8/22/2005 18:28'!
setTraitCompositionFrom: aTraitExpression
	^ self setTraitComposition: aTraitExpression asTraitComposition! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'MarcusDenker 4/26/2012 14:15'!
sourceCodeAt: selector

	^ (self methodDict at: selector) sourceCode.! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'MarcusDenker 4/26/2012 14:15'!
sourceCodeAt: selector ifAbsent: aBlock

	^ (self methodDict at: selector ifAbsent: [^ aBlock value]) sourceCode.! !

!TPureBehavior methodsFor: 'private' stamp: 'nice 1/16/2010 12:52'!
spaceUsed
	"Answer a rough estimate of number of bytes used by this class and its metaclass. Does not include space used by class variables."

	| space |
	space := 0.
	self methodsDo: [:method  |
		space := space + 16.  "dict and org'n space"
		space := space + (method size + 6 "hdr + avg pad").
		method literalsDo: [:lit |
			(lit isMemberOf: Array) ifTrue: [space := space + ((lit size + 1) * 4)].
			(lit isMemberOf: Float) ifTrue: [space := space + 12].
			(lit isMemberOf: ByteString) ifTrue: [space := space + (lit size + 6)].
			(lit isMemberOf: LargeNegativeInteger) ifTrue: [space := space + ((lit size + 1) * 4)].
			(lit isMemberOf: LargePositiveInteger) ifTrue: [space := space + ((lit size + 1) * 4)]]].
		^ space! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:51'!
standardMethodHeaderFor: aSelector
	| args |
	args := (1 to: aSelector numArgs)	collect:[:i| 'arg', i printString].
	args size = 0 ifTrue:[^aSelector asString].
	args size = 1 ifTrue:[^aSelector,' arg1'].
	^String streamContents:[:s|
		(aSelector findTokens:':') with: args do:[:tok :arg|
			s nextPutAll: tok; nextPutAll:': '; nextPutAll: arg; nextPutAll:' '.
		].
	].
! !

!TPureBehavior methodsFor: 'printing' stamp: 'al 12/1/2005 15:37'!
storeLiteral: aCodeLiteral on: aStream
	"Store aCodeLiteral on aStream, changing an Association to ##GlobalName
	 or ###MetaclassSoleInstanceName format if appropriate"
	| key value |
	(aCodeLiteral isVariableBinding)
		ifFalse:
			[aCodeLiteral storeOn: aStream.
			 ^self].
	key := aCodeLiteral key.
	(key isNil and: [(value := aCodeLiteral value) isMemberOf: Metaclass])
		ifTrue:
			[aStream nextPutAll: '###'; nextPutAll: value soleInstance name.
			 ^self].
	(key isSymbol and: [(self bindingOf: key) notNil])
		ifTrue:
			[aStream nextPutAll: '##'; nextPutAll: key.
			 ^self].
	aCodeLiteral storeOn: aStream! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'HenrikSperreJohansen 2/5/2010 14:06'!
thoroughHasSelectorReferringTo: literal special: specialFlag byte: specialByte
	"Answer true if any of my methods access the argument as a 
	literal. Dives into the compact literal notation, making it slow but 
	thorough "

	
	self methodsDo: [ :method |
		((method hasLiteralThorough: literal) or: [specialFlag and: [method scanFor: specialByte]])
			ifTrue: [^true]].
	^false! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'marcus.denker 9/29/2008 08:46'!
thoroughWhichSelectorsReferTo: literal special: specialFlag byte: specialByte
	"Answer a set of selectors whose methods access the argument as a 
	literal. Dives into the compact literal notation, making it slow but 
	thorough "

	| selectors |
	selectors := IdentitySet new.
	self selectorsAndMethodsDo: [:sel :method |
		((method hasLiteralThorough: literal) or: [specialFlag and: [method scanFor: specialByte]])
			ifTrue: [selectors add: sel]].
	^ selectors! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 9/16/2005 14:30'!
traitComposition
	"Return my trait composition. Manipulating the composition does not
	effect changes automatically. Use #setTraitComposition: to do this but
	note, that you have to make a copy of the old trait composition before
	changing it because only the difference between the new and the old
	composition is updated."
	
	^self explicitRequirement ! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 9/16/2005 14:53'!
traitComposition: aTraitComposition
	^self explicitRequirement ! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 9/16/2005 14:31'!
traitCompositionIncludes: aTrait
	^self == aTrait or: 
		[self hasTraitComposition and: 
			[self traitComposition allTraits includes: aTrait]]! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 3/26/2006 21:29'!
traitCompositionString
	^self hasTraitComposition
		ifTrue: [self traitComposition asString]
		ifFalse: ['{}']! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 7/30/2004 21:55'!
traitOrClassOfSelector: aSymbol
	"Return the trait or the class which originally defines the method aSymbol
	or return self if locally defined or if it is a conflict marker method.
	This is primarly used by Debugger to determin the behavior in which a recompiled
	method should be put. If a conflict method is recompiled it should be put into
	the class, thus return self. Also see TraitComposition>>traitProvidingSelector:"
	
	((self includesLocalSelector: aSymbol) or: [
		self hasTraitComposition not]) ifTrue: [^self].
	^(self traitComposition traitProvidingSelector: aSymbol) ifNil: [self]! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 9/16/2005 14:53'!
traitTransformations 
	^ self traitComposition transformations ! !

!TPureBehavior methodsFor: 'traits' stamp: 'damiencassou 1/6/2009 10:58'!
traits
	"Returns a collection of all traits used by the receiver"
	^ self traitComposition traits! !

!TPureBehavior methodsFor: 'traits' stamp: 'MarianoMartinezPeck 6/26/2012 10:40'!
traitsProvidingSelector: aSymbol
	| result |
	result := OrderedCollection new.
	self hasTraitComposition ifFalse: [^result].
	(self traitComposition methodDescriptionsForSelector: aSymbol)
		do: [:methodDescription | methodDescription selector = aSymbol ifTrue: [
			result addAll: (methodDescription locatedMethods
				collect: [:each | each methodClass])]].
	^result! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:51'!
ultimateSourceCodeAt: selector ifAbsent: aBlock
	"Return the source code at selector"
	
	^self
		sourceCodeAt: selector
		ifAbsent: aBlock! !

!TPureBehavior methodsFor: 'traits' stamp: 'GuillermoPolito 7/7/2012 14:34'!
updateMethodDictionarySelector: aSymbol
	"A method with selector aSymbol in myself or my traitComposition has been changed.
	Do the appropriate update to my methodDict (remove or update method) and
	return all affected selectors of me so that my useres get notified."

	| modifiedSelectors descriptions oldProtocol |
	modifiedSelectors := IdentitySet new.
	descriptions := self traitComposition methodDescriptionsForSelector: aSymbol.
	descriptions do: [:methodDescription | | effectiveMethod selector |
		selector := methodDescription selector.
		(self includesLocalSelector: selector) ifFalse: [
			methodDescription isEmpty
				ifTrue: [
					self removeTraitSelector: selector.
					modifiedSelectors add: selector]
				ifFalse: [
					effectiveMethod := methodDescription effectiveMethod.
					self addTraitSelector: selector withMethod: effectiveMethod.
					
					"If the method was not categorized yet, we categorize it "
					oldProtocol := self organization categoryOfElement: selector.
					(oldProtocol isNil or: [ oldProtocol = ClassOrganizer default ]) ifTrue: [ self organization classify: selector under: methodDescription effectiveMethodCategory. ].
					
					modifiedSelectors add: selector]]].
	^modifiedSelectors! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'StephaneDucasse 3/20/2010 22:05'!
whichSelectorsReferTo: literal 
	"Answer a Set of selectors whose methods access the argument as a
literal."

	| special byte |
	special := Smalltalk hasSpecialSelector: literal ifTrueSetByte: [:b |
byte := b].
	^self whichSelectorsReferTo: literal special: special byte: byte

	"Rectangle whichSelectorsReferTo: #+."! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'md 2/15/2006 11:01'!
whichSelectorsReferTo: literal special: specialFlag byte: specialByte
	"Answer a set of selectors whose methods access the argument as a literal."

	| who |
	who := IdentitySet new.
	self selectorsAndMethodsDo: 
		[:sel :method |
		((method hasLiteral: literal) or: [specialFlag and: [method scanFor: specialByte]])
			ifTrue:
				[((literal isVariableBinding) not
					or: [method literals allButLast includes: literal])
						ifTrue: [who add: sel]]].
	^ who! !

!TPureBehavior methodsFor: 'accessing class hierarchy' stamp: 'damiencassou 1/20/2009 10:35'!
withAllSubclassesDo: aBlock
	| temp |
	temp := self allSubclassesDo: aBlock.
	aBlock value: self! !

!TPureBehavior methodsFor: 'accessing class hierarchy' stamp: 'al 12/31/2005 13:40'!
withAllSuperclasses
	"Answer an OrderedCollection of the receiver and the receiver's 
	superclasses. The first element is the receiver, 
	followed by its superclass; the last element is Object."

	| temp |
	temp := self allSuperclasses.
	temp addFirst: self.
	^ temp! !

Trait named: #TTransformationCompatibility
	uses: {}
	category: 'Traits-Kernel-Traits'!

!TTransformationCompatibility commentStamp: 'apb 8/22/2005 15:24' prior: 0!
methods that make the user of this trait protocol-compatible with TraitTransformation.
When used in Trait and ClassTrait, this eliminates the need for TraitHolders.
!

!TTransformationCompatibility methodsFor: 'enquiries' stamp: 'apb 8/22/2005 15:35'!
aliasesForSelector: aSelector 
	^ OrderedCollection new
! !

!TTransformationCompatibility methodsFor: 'enquiries' stamp: 'apb 8/22/2005 15:36'!
allAliasesDict
	^IdentityDictionary new
! !

!TTransformationCompatibility methodsFor: 'enquiries' stamp: 'apb 8/22/2005 15:36'!
changedSelectorsComparedTo: aTraitTransformation
	| selectors otherSelectors changedSelectors aliases otherAliases |
	selectors := self allSelectors asIdentitySet.
	otherSelectors := aTraitTransformation allSelectors asIdentitySet.
	changedSelectors := IdentitySet withAll: (
		(selectors difference: otherSelectors) union: (otherSelectors difference: selectors)).
	aliases := self allAliasesDict.
	otherAliases := aTraitTransformation allAliasesDict.
	aliases keysAndValuesDo: [:key :value |
		(value ~~ (otherAliases at: key ifAbsent: [nil])) ifTrue: [changedSelectors add: key]].
	otherAliases keysAndValuesDo: [:key :value |
		(value ~~ (aliases at: key ifAbsent: [nil])) ifTrue: [changedSelectors add: key]].
	^ changedSelectors.
! !

!TTransformationCompatibility methodsFor: 'enquiries' stamp: 'mada 5/5/2012 11:40'!
collectMethodsFor: aSelector into: methodDescription
	(self includesSelector: aSelector) ifTrue: [ 
		methodDescription addLocatedMethod: 
			(self compiledMethodAt: aSelector)
		]! !

!TTransformationCompatibility methodsFor: 'enquiries' stamp: 'dvf 9/22/2005 18:15'!
subject
	"for compatibility with TraitTransformations"
	^ self
! !

!TTransformationCompatibility methodsFor: 'enquiries' stamp: 'apb 8/22/2005 15:36'!
trait
	"for compatibility with TraitTransformations"
	^ self
! !

!SequenceableCollection methodsFor: '*Traits' stamp: 'apb 8/22/2005 17:07'!
asTraitComposition
	"For convenience the composition {T1. T2 ...} is the same as T1 + T2 + ..."
	^self isEmpty
		ifFalse: [
			self size = 1
				ifTrue: [self first asTraitComposition]
				ifFalse: [
					self copyWithoutFirst 
						inject: self first
						into: [:left :right | left + right]]]
		ifTrue: [
			TraitComposition new]! !

Object subclass: #TraitBehavior
	uses: TPureBehavior @ {#pureRemoveSelector:->#removeSelector:. #pureAddSelectorSilently:withMethod:->#addSelectorSilently:withMethod:}
	instanceVariableNames: 'methodDict traitComposition localSelectors users'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Kernel'!

!TraitBehavior commentStamp: 'apb 3/2/2006 17:49' prior: 0!
I declare the instance variables methodDict, traitComposition and localSelectors that normally would be defined on PureBehavior (see class comment of PureBehavior). Furthermore I keep track of where I am used and I provide compatibility methods to make me useable in tools where originally a class was expected.!

TraitBehavior class
	uses: TPureBehavior classTrait
	instanceVariableNames: ''!

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/20/2005 21:24'!
addSelectorSilently: selector withMethod: compiledMethod
	self pureAddSelectorSilently: selector withMethod: compiledMethod.
	self notifyUsersOfChangedSelector: selector.! !

!TraitBehavior methodsFor: 'traits' stamp: 'CamilloBruni 4/27/2012 18:57'!
addTraitSelector: aSymbol withMethod: aCompiledMethod
	"Add aMethod with selector aSymbol to my
	methodDict. aMethod must not be defined locally."

	| source methodAndNode |
	[(self includesLocalSelector: aSymbol) not] assert.
	self ensureLocalSelectors.
		
	source := aCompiledMethod getSourceReplacingSelectorWith: aSymbol.
	methodAndNode  := self
		compile: source
		classified: aCompiledMethod category
		notifying: nil
		trailer: self defaultMethodTrailer
		ifFail: [^nil].
	methodAndNode method putSource: source fromParseNode: methodAndNode node inFile: 2
		withPreamble: [:f | f cr; nextPut: $!!; nextChunkPut: 'Trait method'; cr].
			
	self basicAddSelector: aSymbol withMethod: methodAndNode method
	"self 
		addAndClassifySelector: aSymbol 
		withMethod: methodAndNode method 
		inProtocol: aCompiledMethod category 
		notifying: nil"! !

!TraitBehavior methodsFor: 'traits' stamp: 'NS 4/12/2004 14:16'!
addUser: aClassOrTrait
	users add: aClassOrTrait! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'dc 5/8/2007 13:51'!
allClassVarNames
	^#()! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'NS 4/12/2004 14:36'!
allInstVarNames
	^ #()! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'nice 12/4/2009 00:07'!
allSelectors
	^ self selectors asSet! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'dc 7/23/2007 11:13'!
allSubclasses
	^ Array new! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'NS 4/12/2004 14:34'!
allSubclassesDo: aBlock! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'al 12/31/2005 13:41'!
allSuperclasses
	^ OrderedCollection new! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'NS 4/12/2004 14:34'!
allSuperclassesDo: aBlock! !

!TraitBehavior methodsFor: 'adding/removing methods' stamp: 'MarcusDenker 2/28/2012 10:07'!
basicAddSelector: selector withMethod: compiledMethod 
	"Add the message selector with the corresponding compiled method to the 
	receiver's method dictionary.
	Do this without sending system change notifications"

	| oldMethodOrNil |	
	oldMethodOrNil := self lookupSelector: selector.
	self methodDict at: selector put: compiledMethod.
	compiledMethod methodClass: self.
	compiledMethod selector: selector.

	"Now flush the method cache, when we replace a method"
	oldMethodOrNil ifNotNil: [oldMethodOrNil flushCache].! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'NS 1/6/2004 09:02'!
basicLocalSelectors
	"Direct accessor for the instance variable localSelectors.
	Since localSelectors is lazily initialized, this may 
	return nil, which means that all selectors are local."

	^ localSelectors! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'NS 3/30/2004 13:26'!
basicLocalSelectors: aSetOrNil
	localSelectors := aSetOrNil! !

!TraitBehavior methodsFor: 'remove me later' stamp: 'apb 8/22/2005 15:28'!
classPool
	^ Dictionary new! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'dc 7/26/2007 16:51'!
classVarNames
	^#()! !

!TraitBehavior methodsFor: 'traits' stamp: 'dvf 9/9/2005 19:39'!
classesComposedWithMe
	^users gather: [:u | u classesComposedWithMe]
! !

!TraitBehavior methodsFor: 'traits' stamp: 'BenjaminVanRyseghem 8/23/2011 18:54'!
hasTraitComposition

	^ traitComposition notNil and: [ traitComposition notEmpty ]! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'al 7/29/2004 15:03'!
inheritsFrom: aClass
	"Used by RB"
	
	^false! !

!TraitBehavior methodsFor: 'initialize-release' stamp: 'alain.plantec 5/28/2009 11:14'!
initialize
	super initialize.
	self methodDict: MethodDictionary new.
	self traitComposition: nil.
	users := IdentitySet new.! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'NS 4/12/2004 14:33'!
instSize
	^0! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'NS 4/12/2004 14:36'!
instVarNames
	^#()! !

!TraitBehavior methodsFor: 'testing' stamp: 'al 2/9/2004 22:29'!
isTrait
	^true! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'al 12/10/2003 10:02'!
lookupSelector: selector
	^(self includesSelector: selector)
		ifTrue: [self compiledMethodAt: selector]
		ifFalse: [nil]! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'NS 12/9/2003 10:51'!
methodDict
	^ methodDict! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'NS 12/9/2003 10:50'!
methodDict: aDictionary
	methodDict := aDictionary! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'damiencassou 2/19/2009 15:09'!
poolDictionaryNames
	^ #()! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/17/2004 14:31'!
precodeCommentOrInheritedCommentFor: aSelector
	^self firstPrecodeCommentFor: aSelector
	! !

!TraitBehavior methodsFor: 'compiling' stamp: 'al 7/30/2004 22:03'!
recompile: selector from: oldClass
	"Compile the method associated with selector in the receiver's method dictionary."
	"ar 7/10/1999: Use oldClass compiledMethodAt: not self compiledMethodAt:"
	| method trailer methodNode |
	method := oldClass compiledMethodAt: selector.
	trailer := method trailer.
	methodNode := self compilerClass new
				compile: (oldClass sourceCodeAt: selector)
				in: self
				notifying: nil
				ifFail: [^ self].   "Assume OK after proceed from SyntaxError"
	selector == methodNode selector ifFalse: [self error: 'selector changed!!'].
	self basicAddSelector: selector withMethod: (methodNode generate: trailer).
! !

!TraitBehavior methodsFor: 'traits' stamp: 'al 4/20/2004 09:33'!
removeFromTraitCompositionOfUsers
	self users do: [:each |
		each removeFromComposition: self ]! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/20/2005 21:24'!
removeSelector: selector
	self pureRemoveSelector: selector.
	self notifyUsersOfChangedSelector: selector.! !

!TraitBehavior methodsFor: 'traits' stamp: 'NS 4/12/2004 14:17'!
removeUser: aClassOrTrait
	users remove: aClassOrTrait ifAbsent: []! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'nice 1/5/2010 15:59'!
selectorsWithArgs: numberOfArgs
	"Return all selectors defined in this class that take this number of arguments.  Could use String.keywords.  Could see how compiler does this."

	| list |
	list := OrderedCollection new.
	self selectorsDo: [:aSel | | num | 
		num := aSel count: [:char | char == $:].
		num = 0 ifTrue: [aSel last isLetter ifFalse: [num := 1]].
		num = numberOfArgs ifTrue: [list add: aSel]].
	^ list! !

!TraitBehavior methodsFor: 'remove me later' stamp: 'StephaneDucasse 12/13/2011 16:34'!
sharedPools
	^ OrderedCollection new! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'NS 4/12/2004 14:34'!
subclassDefinerClass
	^nil subclassDefinerClass ! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'dc 7/23/2007 11:13'!
subclasses
	^ Array new! !

!TraitBehavior methodsFor: 'testing method dictionary' stamp: 'G.C 10/22/2008 09:59'!
thoroughWhichSelectorsReferTo: literal special: specialFlag byte: specialByte 
	"Answer a set of selectors whose methods access the argument as a 
	literal. Dives into the compact literal notation, making it slow but 
	thorough "
	| selectors |
	selectors := IdentitySet new.
	self selectorsAndMethodsDo: 
		[ :sel :method | 
		((method refersToLiteral: literal) or: [ specialFlag and: [ method scanFor: specialByte ] ]) ifTrue: [ selectors add: sel ] ].
	^ selectors! !

!TraitBehavior methodsFor: 'traits' stamp: 'al 12/12/2003 21:42'!
traitComposition
	traitComposition ifNil: [traitComposition := TraitComposition new].
	^traitComposition! !

!TraitBehavior methodsFor: 'traits' stamp: 'NS 12/11/2003 16:35'!
traitComposition: aTraitComposition
	traitComposition := aTraitComposition! !

!TraitBehavior methodsFor: 'traits' stamp: 'mada 5/5/2012 11:33'!
traitsProvidingSelector: aSymbol
	| result |
	result := OrderedCollection new.
	self hasTraitComposition ifFalse: [^result].
	(self traitComposition methodDescriptionsForSelector: aSymbol)
		do: [:methodDescription | methodDescription selector = aSymbol ifTrue: [
			result addAll: (methodDescription locatedMethods
				collect: [:each | each methodClass])]].
	^result! !

!TraitBehavior methodsFor: 'traits' stamp: 'NS 4/12/2004 14:17'!
users
	^users! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'al 7/17/2004 14:35'!
whichClassIncludesSelector: aSymbol 
	"Traits compatibile implementation for:
	
	Answer the class on the receiver's superclass chain where the 
	argument, aSymbol (a message selector), will be found. Answer nil if none found."
	
	^(self includesSelector: aSymbol)
		ifTrue: [self]
		ifFalse: [nil]! !

!TraitBehavior methodsFor: 'testing method dictionary' stamp: 'StephaneDucasse 3/20/2010 22:05'!
whichSelectorsReferTo: literal 
	"Answer a Set of selectors whose methods access the argument as a
literal."

	| special byte |
	special := Smalltalk hasSpecialSelector: literal ifTrueSetByte: [:b |
byte := b].
	^self whichSelectorsReferTo: literal special: special byte: byte

	"Rectangle whichSelectorsReferTo: #+."! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'damiencassou 3/13/2009 19:27'!
withAllSubclasses
	^ Array with: self! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'al 2/8/2004 19:08'!
zapAllMethods
	"Remove all methods in this trait which is assumed to be obsolete"

	methodDict := MethodDictionary new.
	self hasClassTrait ifTrue: [self classTrait zapAllMethods]! !

TraitBehavior subclass: #TraitDescription
	uses: TClassAndTraitDescription + TComposingDescription + TTransformationCompatibility
	instanceVariableNames: 'organization'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Kernel'!

!TraitDescription commentStamp: '<historical>' prior: 0!
I add a number of facilities (most defined by traits which are also used in ClassDescription):
	Category organization for methods
	The maintenance of a ChangeSet, and logging changes on a file
	Most of the mechanism for fileOut.
	Copying of methods to other traits/classes
	Operators to create trait compositions!

TraitDescription class
	uses: TComposingDescription classTrait + TTransformationCompatibility classTrait + TClassAndTraitDescription classTrait
	instanceVariableNames: ''!

TraitDescription subclass: #ClassTrait
	uses: TApplyingOnClassSide
	instanceVariableNames: 'baseTrait'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Kernel'!

!ClassTrait commentStamp: '<historical>' prior: 0!
While every class has an associated metaclass, a trait can have an associated classtrait, an instance of me. To preserve metaclass compatibility, the associated classtrait (if there is one) is automatically applied to the metaclass, whenever a trait is applied to a class. Consequently, a trait with an associated classtrait can only be applied to classes, whereas a trait without a classtrait can be applied to both classes and metaclasses.!

ClassTrait class
	uses: TApplyingOnClassSide classTrait
	instanceVariableNames: ''!

!ClassTrait class methodsFor: 'instance creation' stamp: 'al 3/23/2004 19:41'!
for: aTrait
	^self new
		initializeWithBaseTrait: aTrait;
		yourself! !

!ClassTrait methodsFor: 'initialize' stamp: 'al 7/18/2004 12:11'!
baseClass: aTrait traitComposition: aComposition methodDict: aMethodDict localSelectors: aSet organization: aClassOrganization

	self baseTrait: aTrait.
	self	
		traitComposition: aComposition
		methodDict: aMethodDict
		localSelectors: aSet
		organization: aClassOrganization
! !

!ClassTrait methodsFor: 'accessing parallel hierarchy' stamp: 'al 3/16/2004 10:38'!
baseTrait
	^baseTrait! !

!ClassTrait methodsFor: 'accessing parallel hierarchy' stamp: 'jannik.laval 5/1/2010 16:02'!
baseTrait: aTrait
	[aTrait isBaseTrait] assert.
	baseTrait := aTrait
	
	! !

!ClassTrait methodsFor: 'accessing' stamp: 'damiencassou 8/6/2009 11:37'!
category
	^ self baseTrait category! !

!ClassTrait methodsFor: 'accessing parallel hierarchy' stamp: 'al 3/16/2004 10:38'!
classTrait
	^self! !

!ClassTrait methodsFor: 'accessing parallel hierarchy' stamp: 'al 4/20/2004 09:44'!
classTrait: aClassTrait
	self error: 'Trait is already a class trait!!'
	
	! !

!ClassTrait methodsFor: 'compiling' stamp: 'CamilloBruni 5/30/2012 12:41'!
compile: text classified: category withStamp: changeStamp notifying: requestor logSource: logSource
	
	| classSideUsersOfBaseTrait message |
	
	classSideUsersOfBaseTrait := self baseTrait users select: [:each | each isClassSide].
	
	classSideUsersOfBaseTrait isEmpty ifFalse: [
		message := String streamContents: [ :stream |
			stream nextPutAll: 'The instance side of this trait is used on '; cr.
			classSideUsersOfBaseTrait
				do: [:each | stream nextPutAll: each name ]
				separatedBy: [ stream nextPutAll: ', ' ].
			stream cr; nextPutAll: ' You can not add methods to the class side of this trait!!'].
		^ TraitException signal:  message ].
	
	^ super
		compile: text
		classified: category
		withStamp: changeStamp
		notifying: requestor
		logSource: logSource! !

!ClassTrait methodsFor: 'copying' stamp: 'dvf 8/30/2005 16:51'!
copy
	"Make a copy of the receiver. Share the 
	reference to the base trait."

	^(self class new)
		baseTrait: self baseTrait;
		initializeFrom: self;
		yourself! !

!ClassTrait methodsFor: 'filein/out' stamp: 'MarcusDenker 2/14/2010 22:18'!
definition
	^String streamContents: [:stream |
		stream
			nextPutAll: self name;
			crtab;
			nextPutAll: 'uses: ';
			nextPutAll: self traitCompositionString]! !

!ClassTrait methodsFor: 'accessing parallel hierarchy' stamp: 'al 3/16/2004 10:41'!
hasClassTrait
	^false! !

!ClassTrait methodsFor: 'initialize' stamp: 'dvf 8/30/2005 16:48'!
initializeFrom: anotherClassTrait
	traitComposition := self traitComposition copyTraitExpression.
	methodDict := self methodDict copy.
	localSelectors := self localSelectors copy.
	organization := self organization copy.! !

!ClassTrait methodsFor: 'initialize' stamp: 'al 3/24/2004 20:37'!
initializeWithBaseTrait: aTrait
	self baseTrait: aTrait.
	self noteNewBaseTraitCompositionApplied: aTrait traitComposition.
	aTrait users do: [:each | self addUser: each classSide].
	! !

!ClassTrait methodsFor: 'accessing parallel hierarchy' stamp: 'al 3/16/2004 10:48'!
isBaseTrait
	^false! !

!ClassTrait methodsFor: 'accessing parallel hierarchy' stamp: 'al 3/16/2004 10:48'!
isClassTrait
	^true! !

!ClassTrait methodsFor: 'testing' stamp: 'nice 11/5/2009 21:51'!
isSelfEvaluating
	"Return true if the receiver printString is evaluating back to self."
	
	^baseTrait isObsolete not! !

!ClassTrait methodsFor: 'accessing' stamp: 'al 4/21/2004 09:38'!
name
	^self baseTrait name , ' classTrait'! !

!ClassTrait methodsFor: 'initialize' stamp: 'al 7/17/2004 22:56'!
traitComposition: aComposition methodDict: aMethodDict localSelectors: aSet organization: aClassOrganization

	"Used by copy of Trait"

	localSelectors := aSet.
	methodDict := aMethodDict.
	traitComposition := aComposition.
	self organization: aClassOrganization! !

!ClassTrait methodsFor: 'composition' stamp: 'EstebanLorenzano 7/27/2012 16:30'!
uses: aTraitCompositionOrArray
	| copyOfOldTrait newComposition |
	copyOfOldTrait := self copy.
	newComposition := aTraitCompositionOrArray asTraitComposition.
	self assertConsistantCompositionsForNew: newComposition.
	self setTraitComposition: newComposition.
	SystemAnnouncer uniqueInstance
		traitDefinitionChangedFrom: copyOfOldTrait to: self.! !

TraitDescription subclass: #Trait
	uses: TBehaviorCategorization
	instanceVariableNames: 'name environment classTrait category'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Kernel'!

!Trait commentStamp: '<historical>' prior: 0!
Each trait in the system is represented as an instance of me. Like Class, I concretize my superclass by providing instance variables for the name and the environment. Since traits do not define variables, I not provide facilities for pool variables. However, I declare an instance variable to hold the associated classtrait, which is an instance of the class ClassTrait. !

Trait class
	uses: TBehaviorCategorization classTrait
	instanceVariableNames: ''!

!Trait class methodsFor: 'instance creation' stamp: 'al 12/7/2003 16:00'!
defaultEnvironment
	^Smalltalk! !

!Trait class methodsFor: 'instance creation' stamp: 'Alexandre Bergel 5/22/2010 14:41'!
named: aSymbol
	^ self named: aSymbol uses: {} category: 'Unclassified'! !

!Trait class methodsFor: 'instance creation' stamp: 'md 3/3/2006 11:05'!
named: aSymbol uses: aTraitCompositionOrCollection category: aString
	| env |
	env := self environment.
	^self
		named: aSymbol
		uses: aTraitCompositionOrCollection
		category: aString
		env: env! !

!Trait class methodsFor: 'instance creation' stamp: 'EstebanLorenzano 11/8/2012 20:27'!
named: aSymbol uses: aTraitCompositionOrCollection category: aString env: anEnvironment
	| trait oldTrait systemCategory oldCategory |
	
	systemCategory := aString asSymbol.
	trait := anEnvironment
		at: aSymbol
		ifAbsent: [nil].
	oldTrait := trait copy.
	"Since there some strange side effect who changes teh category before being able to 
	  actually update (and that causes a problem on RPackage organization), we take also the 
	  category"
	oldCategory := oldTrait ifNotNil: [ oldTrait category ].
	trait := trait ifNil: [ super new ].
	
	(trait isKindOf: Trait) ifFalse: [
		^self error: trait name , ' is not a Trait'].
	trait
		setName: aSymbol
		andRegisterInCategory: systemCategory
		environment: anEnvironment.
		
	trait setTraitComposition: aTraitCompositionOrCollection asTraitComposition.
	
	"... notify interested clients ..."
	oldTrait ifNil: [
		SystemAnnouncer uniqueInstance 
			classAdded: trait 
			inCategory: systemCategory.
		^trait].

	SystemAnnouncer uniqueInstance traitDefinitionChangedFrom: oldTrait to: trait.

	systemCategory ~= oldCategory
		ifTrue: [ 
			SystemAnnouncer uniqueInstance 
				class: trait 
				recategorizedFrom: oldCategory 
				to: systemCategory].
		
	^ trait! !

!Trait class methodsFor: 'instance creation' stamp: 'al 12/19/2003 15:03'!
new
	self shouldNotImplement! !

!Trait class methodsFor: 'printing' stamp: 'al 2/9/2004 22:23'!
newTemplateIn: categoryString
	^String streamContents: [:stream |
		stream
			nextPutAll: self name;
			nextPutAll: ' named: #NameOfTrait';
			cr; tab;
			nextPutAll: 'uses: {}';
			cr; tab;
			nextPutAll: 'category: ';
			nextPut: $';
			nextPutAll: categoryString;
			nextPut: $' ]! !

!Trait methodsFor: 'accessing class hierarchy' stamp: 'MarcusDenker 6/12/2010 15:32'!
allSubclassesWithLevelDo: classAndLevelBlock startingLevel: level 
	"Walk the tree of subclasses, giving the class and its level"
	"Empty implementation needed, compatibility with OmniBrowser"
! !

!Trait methodsFor: 'private' stamp: 'CamilleTeruel 2/11/2013 14:35'!
applyChangesOfNewTraitCompositionReplacing: oldComposition
	"Duplicated on Class"
	
	| changedSelectors |
	changedSelectors := super applyChangesOfNewTraitCompositionReplacing: oldComposition.
	self classSide noteNewBaseTraitCompositionApplied: self traitComposition.
	^ changedSelectors! !

!Trait methodsFor: 'accessing parallel hierarchy' stamp: 'al 7/17/2004 14:16'!
baseTrait
	^self! !

!Trait methodsFor: 'accessing' stamp: 'al 3/18/2006 13:23'!
basicCategory
	^category! !

!Trait methodsFor: 'accessing' stamp: 'al 3/18/2006 13:22'!
basicCategory: aSymbol
	category := aSymbol! !

!Trait methodsFor: 'compiling' stamp: 'MarcusDenker 8/15/2010 19:16'!
binding
	^ self environment associationAt: name ifAbsent: [ nil -> self ]! !

!Trait methodsFor: 'accessing parallel hierarchy' stamp: 'al 7/17/2004 14:16'!
classTrait
	^classTrait! !

!Trait methodsFor: 'accessing parallel hierarchy' stamp: 'jannik.laval 5/1/2010 16:36'!
classTrait: aTrait
	"Assigns the class trait associated with the receiver."
	
	[aTrait isClassTrait] assert.
	classTrait := aTrait! !

!Trait methodsFor: 'copying' stamp: 'MarcusDenker 8/16/2010 15:28'!
copy 
	| newTrait |
	newTrait := self class basicNew initialize
		name: self name
		traitComposition: self traitComposition copyTraitExpression 
		methodDict: self methodDict copy
		localSelectors: self localSelectors copy
		organization: self organization copy.
	
	newTrait environment: self environment.	
	newTrait classTrait initializeFrom: self classTrait.
	^newTrait! !

!Trait methodsFor: 'copying' stamp: 'BenjaminVanRyseghem 12/16/2011 10:45'!
duplicateClassWithNewName: aSymbol
	" I am duplicated from Class, and used to duplicate a trait "

	| copysName class newDefinition |

	copysName := aSymbol asSymbol.
	copysName = self name
		ifTrue: [ ^ self ].
	(Smalltalk globals includesKey: copysName)
		ifTrue: [ ^ self error: copysName , ' already exists' ].
	newDefinition := self definition copyReplaceAll: '#' , self name asString with: '#' , copysName asString.
	class := self class evaluatorClass evaluate: newDefinition logged: true.
	
	class copyAllCategoriesFrom: self.
	class class copyAllCategoriesFrom: self class.
	^ class! !

!Trait methodsFor: 'accessing' stamp: 'al 7/17/2004 14:16'!
environment
	^environment! !

!Trait methodsFor: 'accessing' stamp: 'al 7/17/2004 14:16'!
environment: anObject
	environment := anObject! !

!Trait methodsFor: 'filein/out' stamp: 'al 12/1/2005 16:29'!
fileOutOn: aFileStream moveSource: moveSource toFile: fileIndex 
	"This is just copied from Class.. the whole fileout is a mess."
	^self fileOutOn: aFileStream moveSource: moveSource toFile: fileIndex initializing: true! !

!Trait methodsFor: 'filein/out' stamp: 'StephaneDucasse 11/7/2011 22:45'!
fileOutOn: aFileStream moveSource: moveSource toFile: fileIndex initializing: aBool
	"File a description of the receiver on aFileStream. If the boolean argument,
	moveSource, is true, then set the trailing bytes to the position of aFileStream and
	to fileIndex in order to indicate where to find the source code."

	self crTrace: name.
	super
		fileOutOn: aFileStream
		moveSource: moveSource
		toFile: fileIndex.
	self hasClassTrait ifTrue: [
		aFileStream cr; nextPutAll: '"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!!'; cr; cr.
		self classTrait
			fileOutOn: aFileStream
			moveSource: moveSource
			toFile: fileIndex]! !

!Trait methodsFor: 'accessing parallel hierarchy' stamp: 'al 7/17/2004 14:16'!
hasClassTrait
	^classTrait notNil! !

!Trait methodsFor: 'initialization' stamp: 'al 7/17/2004 14:16'!
initialize
	super initialize.
	classTrait := ClassTrait for: self! !

!Trait methodsFor: 'accessing parallel hierarchy' stamp: 'al 7/17/2004 14:16'!
isBaseTrait
	^true! !

!Trait methodsFor: 'accessing parallel hierarchy' stamp: 'al 7/17/2004 14:16'!
isClassTrait
	^false! !

!Trait methodsFor: 'testing' stamp: 'al 7/17/2004 14:16'!
isObsolete
	"Return true if the receiver is obsolete."
	^(self environment at: name ifAbsent: [nil]) ~~ self! !

!Trait methodsFor: 'testing' stamp: 'nice 11/5/2009 21:49'!
isSelfEvaluating
	"Return true if the receiver printString is evaluating back to self."
	
	^self isObsolete not! !

!Trait methodsFor: 'private' stamp: 'jannik.laval 2/5/2010 21:40'!
isValidTraitName: aSymbol
	^(aSymbol isEmptyOrNil
		or: [aSymbol first isLetter not
		or: [aSymbol anySatisfy: [:character | character isAlphaNumeric not]]]) not! !

!Trait methodsFor: 'accessing' stamp: 'al 7/17/2004 14:16'!
name
	^name! !

!Trait methodsFor: 'accessing' stamp: 'al 7/17/2004 14:16'!
name: aSymbol
	name := aSymbol! !

!Trait methodsFor: 'initialization' stamp: 'al 7/17/2004 23:00'!
name: aString traitComposition: aComposition methodDict: aMethodDict localSelectors: aSet organization: aClassOrganization

	"Used by copy"
	
	self name: aString.
	localSelectors := aSet.
	methodDict := aMethodDict.
	traitComposition := aComposition.
	self organization: aClassOrganization
	
	! !

!Trait methodsFor: 'initialization' stamp: 'al 7/29/2004 16:48'!
obsolete
	self name: ('AnObsolete' , self name) asSymbol.
	self hasClassTrait ifTrue: [
		self classTrait obsolete].
	super obsolete! !

!Trait methodsFor: 'compiling' stamp: 'md 7/13/2012 17:00'!
possibleVariablesFor: misspelled continuedFrom: oldResults

	| results |
	results := misspelled correctAgainstDictionary: self classPool continuedFrom: oldResults.
	^ misspelled correctAgainstDictionary: self environment continuedFrom: results.
		! !

!Trait methodsFor: 'initialization' stamp: 'al 7/17/2004 14:16'!
removeFromSystem
	self removeFromSystem: true! !

!Trait methodsFor: 'initialization' stamp: 'CamilleTeruel 2/12/2013 10:55'!
removeFromSystem: logged
	self users do: [ :user | user removeFromComposition: self ].
	self environment forgetClass: self logged: logged.
	self obsolete! !

!Trait methodsFor: 'private' stamp: 'al 7/17/2004 14:16'!
rename: aString 
	"The new name of the receiver is the argument, aString."

	| newName |
	(newName := aString asSymbol) ~= self name
		ifFalse: [^ self].
	(self environment includesKey: newName)
		ifTrue: [^ self error: newName , ' already exists'].
	(Undeclared includesKey: newName)
		ifTrue: [self inform: 'There are references to, ' , aString printString , '
from Undeclared. Check them after this change.'].
	self environment renameClass: self as: newName.
	name := newName! !

!Trait methodsFor: 'private' stamp: 'al 7/17/2004 14:16'!
setName: aSymbol andRegisterInCategory: categorySymbol environment: aSystemDictionary
	(self isValidTraitName: aSymbol) ifFalse: [TraitException signal: 'Invalid trait name'].
	
	(self environment == aSystemDictionary
		and: [self name = aSymbol
			and: [self category = categorySymbol]]) ifTrue: [^self].
		
	((aSystemDictionary includes: aSymbol) and: [(aSystemDictionary at: aSymbol) ~~ self])
		ifTrue: [TraitException signal: 'The name ''' , aSymbol , ''' is already used'].

	(self environment notNil and: [self name notNil and: [self name ~= aSymbol]]) ifTrue: [
		self environment renameClass: self as: aSymbol].
	
	self name: aSymbol.
	self environment: aSystemDictionary.	
	self environment at: self name put: self.
	self environment organization classify: self name under: categorySymbol.
	^ true! !

!Trait methodsFor: 'compatibility' stamp: 'StephaneDucasse 3/18/2011 09:27'!
subclasses
	"for compatibility with classes"
	
	^ #()! !

!Trait methodsFor: 'compatibility' stamp: 'StephaneDucasse 3/18/2011 09:27'!
superclass
	"For compatibility with classes"

	^ nil! !

!Trait methodsFor: 'compiling' stamp: 'eem 7/18/2008 18:05'!
variablesAndOffsetsDo: aBinaryBlock
	"This is the interface between the compiler and a class's instance or field names.  The
	 class should enumerate aBinaryBlock with the field definitions (with nil offsets) followed
	 by the instance variable name strings and their integer offsets (1-relative).  The order is
	 important; names evaluated later will override the same names occurring earlier."

	"Since Traits don't confer state there is nothing to do here."! !

!TraitDescription methodsFor: 'composition' stamp: 'CamilloBruni 4/27/2012 16:58'!
- anArrayOfSelectors
	^TraitExclusion
		with: self
		exclusions: (anArrayOfSelectors isSymbol 
							ifTrue: [{ anArrayOfSelectors }] 
							ifFalse: [ anArrayOfSelectors])! !

!TraitDescription methodsFor: 'accessing method dictionary' stamp: 'EstebanLorenzano 8/3/2012 15:14'!
addAndClassifySelector: selector withMethod: compiledMethod inProtocol: category notifying: requestor 
	| priorMethodOrNil oldProtocol newProtocol |
	
	priorMethodOrNil := self
				compiledMethodAt: selector
				ifAbsent: [].
	
	"We categorize first, so the users can use the categorization at the add moment"			
	oldProtocol := self organization categoryOfElement: selector.
	SystemAnnouncer uniqueInstance
		suspendAllWhile: [ self organization classify: selector under: category ].
	
	self addSelectorSilently: selector withMethod: compiledMethod.
	
	newProtocol := self organization categoryOfElement: selector.
	priorMethodOrNil isNil
		ifTrue: [
			SystemAnnouncer uniqueInstance
				methodAdded: compiledMethod
				selector: selector
				inProtocol: category
				class: self
				requestor: requestor]
		ifFalse: [
			SystemAnnouncer uniqueInstance
				methodChangedFrom: priorMethodOrNil
				to: compiledMethod
				selector: selector
				inClass: self
				oldProtocol: oldProtocol
				newProtocol: newProtocol
				requestor: requestor]! !

!TraitDescription methodsFor: 'composition' stamp: 'al 12/7/2005 21:00'!
addExclusionOf: aSymbol
	^self - {aSymbol}! !

!TraitDescription methodsFor: 'accessing method dictionary' stamp: 'MarcusDenker 3/5/2010 14:36'!
allMethodsInCategory: aName 
	"Answer a list of all the method categories of the receiver"
	
	| aColl |
	aColl := aName = ClassOrganizer allCategory
		ifTrue: [self organization allMethodSelectors]
		ifFalse: [self organization listAtCategoryNamed: aName].
	^aColl asSet asArray sort! !

!TraitDescription methodsFor: 'accessing method dictionary' stamp: 'CamilloBruni 10/20/2012 23:11'!
allSelectorsWithout: aBehaviorsCollection
	"For compatibility"
	^ self selectors! !

!TraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'al 4/21/2004 10:01'!
baseTrait
	self subclassResponsibility! !

!TraitDescription methodsFor: 'organization' stamp: 'al 3/18/2006 14:10'!
category
	self subclassResponsibility! !

!TraitDescription methodsFor: 'organization' stamp: 'al 3/18/2006 14:10'!
category: aString
	self subclassResponsibility! !

!TraitDescription methodsFor: 'accessing comment' stamp: 'PeterHugossonMiller 9/3/2009 11:45'!
classCommentBlank

	| existingComment stream |
	existingComment := self instanceSide organization classComment.
	existingComment isEmpty
		ifFalse: [^existingComment].

	stream := (String new: 100) writeStream.
	stream
		nextPutAll: 'A';
		nextPutAll: (self name first isVowel ifTrue: ['n '] ifFalse: [' ']);
		nextPutAll: self name;
		nextPutAll: ' is xxxxxxxxx.'.
	stream cr.
	^stream contents! !

!TraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'NS 4/12/2004 15:05'!
classSide
	^self classTrait! !

!TraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'al 4/21/2004 10:01'!
classTrait
	self subclassResponsibility! !

!TraitDescription methodsFor: 'enquiries' stamp: 'mada 5/5/2012 11:36'!
collectMethodsFor: aSelector into: methodDescription
	(self includesSelector: aSelector) ifTrue: [ 
		methodDescription addLocatedMethod: 
			(self compiledMethodAt: aSelector)
		]! !

!TraitDescription methodsFor: 'compiling' stamp: 'CamilloBruni 5/30/2012 12:41'!
compile: text classified: category notifying: requestor
	| stamp |
	stamp := self acceptsLoggingOfCompilation 
		ifTrue: [ Author changeStamp ] 
		ifFalse: [ nil ].
	^ self 
		compile: text 
		classified: category
		withStamp: stamp 
		notifying: requestor! !

!TraitDescription methodsFor: 'compiling' stamp: 'CamilloBruni 5/30/2012 11:28'!
compile: text classified: category withStamp: changeStamp notifying: requestor
	^ self 
		compile: text 
		classified: category 
		withStamp: changeStamp 
		notifying: requestor 
		logSource: self acceptsLoggingOfCompilation! !

!TraitDescription methodsFor: 'compiling' stamp: 'ClementBera 2/8/2013 15:14'!
compile: text classified: category withStamp: changeStamp notifying: requestor logSource: logSource

	| methodAndNode |

	methodAndNode := self
		compile: text asString 
		classified: category
		notifying: requestor
		trailer: self defaultMethodTrailer 
		ifFail: [ ^ nil ].
		
	logSource ifTrue: [
		self 
			logMethodSource: text asString
			forMethodWithNode: methodAndNode
			inCategory: category 
			withStamp: changeStamp 
			notifying: requestor ].
		
	self 
		addAndClassifySelector: methodAndNode selector 
		withMethod: methodAndNode method 
		inProtocol: category 
		notifying: requestor.
			
	self instanceSide 
		noteCompilationOf: methodAndNode selector 
		meta: self isClassSide.

	^ methodAndNode selector! !

!TraitDescription methodsFor: 'copying' stamp: 'dvf 9/22/2005 17:54'!
copyTraitExpression
	"When recursively copying a trait expression, the primitive traits should NOT be copied
because they are globally named 'well-known' objects"
	^ self ! !

!TraitDescription methodsFor: 'fileIn/Out' stamp: 'PeterHugossonMiller 9/3/2009 11:45'!
fileOut
	"Create a file whose name is the name of the receiver with '.st' as the 
	extension, and file a description of the receiver onto it."
	
	| internalStream |
	internalStream := (String new: 100) writeStream.
	internalStream header; timeStamp.

	self fileOutOn: internalStream moveSource: false toFile: 0.
	internalStream trailer.

	FileStream writeSourceCodeFrom: internalStream baseName: self name isSt: true! !

!TraitDescription methodsFor: 'filein/out' stamp: 'nice 1/5/2010 15:59'!
fileOutChangedMessages: aSet on: aFileStream moveSource: moveSource toFile: fileIndex 
	"File a description of the messages of this class that have been 
	changed (i.e., are entered into the argument, aSet) onto aFileStream.  If 
	moveSource, is true, then set the method source pointer to the new file position.
	Note when this method is called with moveSource=true, it is condensing the
	.changes file, and should only write a preamble for every method."
	| org |
	(org := self organization) categories do: 
		[:cat | | sels | 
		sels := (org listAtCategoryNamed: cat) select: [:sel | aSet includes: sel].
		sels do:
			[:sel |  self printMethodChunk: sel withPreamble: true on: aFileStream
							moveSource: moveSource toFile: fileIndex]]! !

!TraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'al 4/21/2004 10:02'!
hasClassTrait
	self subclassResponsibility! !

!TraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'NS 4/12/2004 15:05'!
instanceSide
	^self baseTrait! !

!TraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'al 3/16/2004 10:47'!
isBaseTrait
	self subclassResponsibility! !

!TraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'al 3/16/2004 10:47'!
isClassTrait
	self subclassResponsibility! !

!TraitDescription methodsFor: 'testing' stamp: 'al 4/9/2004 15:28'!
isTestCase
	^false! !

!TraitDescription methodsFor: 'private' stamp: 'StephaneDucasse 3/3/2010 13:42'!
linesOfCode 
	"An approximate measure of lines of code.
	Includes comments, but excludes blank lines."
	| lines |
	lines := self localMethods inject: 0 into: [:sum :each | sum + each linesOfCode]. 
	^ self isMeta 
		ifTrue: [lines]
		ifFalse: [lines + self class linesOfCode]! !

!TraitDescription methodsFor: 'private' stamp: 'alain.plantec 5/18/2009 08:41'!
logMethodSource: aText forMethodWithNode: aCompiledMethodWithNode inCategory: category withStamp: changeStamp notifying: requestor 
	aCompiledMethodWithNode method
		putSource: aText
		fromParseNode: aCompiledMethodWithNode node
		class: self
		category: category
		withStamp: changeStamp 
		inFile: 2
		priorMethod: (self
				compiledMethodAt: aCompiledMethodWithNode selector
				ifAbsent: [])! !

!TraitDescription methodsFor: 'accessing method dictionary' stamp: ''!
methodsInCategory: aName 
	"Answer a list of the methods of the receiver that are in category named aName"
	
	| aColl |
	aColl := Set withAll: (aName = ClassOrganizer allCategory
			ifTrue: [self organization allMethodSelectors]
			ifFalse: [self organization listAtCategoryNamed: aName]).
	^ aColl asArray sort! !

!TraitDescription methodsFor: 'organization updating' stamp: 'CamilloBruni 4/27/2012 14:53'!
noteChangesFrom: oldMethodDict
	"create notifications about the changes made to the method dictionary"
	
	"deal with additions"
	self methodDict keysAndValuesDo:  [ :selector :currentMethod|
		oldMethodDict 
			at: selector 
			ifAbsent: [ self noteMethodAdded: currentMethod ]].
	
	" deal with removal / updates"
	oldMethodDict keysAndValuesDo: [ :selector :oldMethod|
		self methodDict 
			at: selector
			ifPresent: [ :currentMethod|
				currentMethod = oldMethod
					ifFalse: [ self noteMethodChanged: oldMethod to: currentMethod]]
			ifAbsent: [ self noteMethodRemoved: oldMethod ]].
	! !

!TraitDescription methodsFor: 'organization updating' stamp: 'EstebanLorenzano 7/27/2012 16:30'!
noteMethodAdded: aMethod
	SystemAnnouncer uniqueInstance
		methodAdded: aMethod! !

!TraitDescription methodsFor: 'organization updating' stamp: 'EstebanLorenzano 7/27/2012 16:30'!
noteMethodChanged: oldMethod to: newMethod
	SystemAnnouncer uniqueInstance
		methodChangedFrom: oldMethod to: newMethod ! !

!TraitDescription methodsFor: 'organization updating' stamp: 'EstebanLorenzano 7/27/2012 16:30'!
noteMethodRemoved: oldMethod
	SystemAnnouncer uniqueInstance
		methodRemoved: oldMethod! !

!TraitDescription methodsFor: 'organization updating' stamp: 'nice 1/5/2010 15:59'!
noteRecategorizedSelectors: aCollection oldComposition: aTraitComposition
	
	aCollection do: [:each | | oldCategory newCategory | 
		oldCategory := self organization categoryOfElement: each.
		newCategory := (self traitComposition methodDescriptionForSelector: each) effectiveMethodCategory.
		self noteRecategorizedSelector: each from: oldCategory to: newCategory]! !

!TraitDescription methodsFor: 'organization updating' stamp: 'EstebanLorenzano 8/3/2012 15:30'!
notifyOfRecategorizedSelector: element from: oldCategory to: newCategory
	SystemAnnouncer uniqueInstance selector: element recategorizedFrom: oldCategory to: newCategory inClass: self.
	SystemAnnouncer uniqueInstance 
		suspendAllWhile: [self notifyUsersOfRecategorizedSelector: element from: oldCategory to: newCategory].! !

!TraitDescription methodsFor: 'users notification' stamp: 'NS 4/12/2004 15:45'!
notifyUsersOfChangedSelectors: aCollection
	self users do: [:each |
		each noteChangedSelectors: aCollection]! !

!TraitDescription methodsFor: 'users notification' stamp: 'NS 4/12/2004 21:46'!
notifyUsersOfRecategorizedSelector: element from: oldCategory to: newCategory
	self users do: [:each |
		each noteRecategorizedSelector: element from: oldCategory to: newCategory]! !

!TraitDescription methodsFor: 'private' stamp: 'MarcusDenker 2/21/2010 12:57'!
numberOfMethods
	"cound all methods that are local (not comming from a trait)"
	| num |
	num := self localMethods size.
	self isMeta 
		ifTrue: [^ num] 
		ifFalse: [^ num + self class numberOfMethods]! !

!TraitDescription methodsFor: 'initialization' stamp: 'NS 4/8/2004 11:01'!
obsolete
	"Make the receiver obsolete."
	self organization: nil.
	super obsolete! !

!TraitDescription methodsFor: 'organization' stamp: 'MarcusDenker 7/13/2012 16:52'!
organization
	"Answer the instance of ClassOrganizer that represents the organization 
	of the messages of the receiver."

	organization ifNil:
		[self organization: (ClassOrganizer defaultList: self selectors asArray sort)].
	
	"Making sure that subject is set correctly. It should not be necessary."
	organization ifNotNil: [organization setSubject: self].
	^ organization! !

!TraitDescription methodsFor: 'organization' stamp: 'NS 4/8/2004 11:04'!
organization: aClassOrg
	"Install an instance of ClassOrganizer that represents the organization of the messages of the receiver."

	aClassOrg ifNotNil: [aClassOrg setSubject: self].
	organization := aClassOrg! !

!TraitDescription methodsFor: 'filein/out' stamp: 'JB 2/12/2010 15:19'!
printCategoryChunk: category on: aFileStream priorMethod: priorMethod
	^ self printCategoryChunk: category on: aFileStream
		withStamp: Author changeStamp priorMethod: priorMethod! !

!TraitDescription methodsFor: 'filein/out' stamp: 'StephaneDucasse 5/28/2011 13:42'!
putClassCommentToCondensedChangesFile: aFileStream
	"Called when condensing changes.  If the receiver has a class comment, and if that class comment does not reside in the .sources file, then write it to the given filestream, with the resulting RemoteString being reachable from the source file #2.  Note that any existing backpointer into the .sources file is lost by this process -- a situation that maybe should be fixed someday."

	| header aStamp aCommentRemoteStr |
	self isMeta ifTrue: [^ self].  "bulletproofing only"
	((aCommentRemoteStr := self organization commentRemoteStr) isNil or:
		[aCommentRemoteStr sourceFileNumber = 1]) ifTrue: [^ self].

	aFileStream cr; nextPut: $!!.
	header := String streamContents: [:strm | strm nextPutAll: self name;
		nextPutAll: ' commentStamp: '.
		(aStamp := self organization commentStamp ifNil: ['<historical>']) storeOn: strm.
		strm nextPutAll: ' prior: 0'].
	aFileStream nextChunkPut: header.
	aFileStream cr.
	self organization classComment: (RemoteString newString: self organization classComment onFileNumber: 2 toFile: aFileStream) stamp: aStamp! !

!TraitDescription methodsFor: 'private' stamp: 'al 7/30/2004 11:49'!
spaceUsed
	^super spaceUsed + (self hasClassTrait
		ifTrue: [self classTrait spaceUsed] 
		ifFalse: [0])! !

!TraitDescription methodsFor: 'class compatibility' stamp: 'NS 12/9/2003 11:18'!
theMetaClass
	^ self classTrait! !

!TraitDescription methodsFor: 'class compatibility' stamp: 'al 12/10/2003 00:07'!
theNonMetaClass
	^ self baseTrait! !

!TraitDescription methodsFor: 'accessing' stamp: 'CamilloBruni 8/1/2012 16:18'!
traitVersion
	"Default.  Any class may return a later version to inform readers that use ReferenceStream. 
	This method allows you to distinguish between class versions when the shape of the class 
	hasn't changed (when there's no change in the instVar names).
	In the conversion methods you usually can tell by the inst var names 
	what old version you have. In a few cases, though, the same inst var 
	names were kept but their interpretation changed (like in the layoutFrame).
	By changing the class version when you keep the same instVars you can 
	warn older and newer images that they have to convert."
	^ 0! !

!TraitDescription methodsFor: 'closure support' stamp: 'sd 3/22/2009 21:59'!
variablesAndOffsetsDo: aBlock

	^self! !

!TraitDescription methodsFor: 'accessing' stamp: 'NS 12/9/2003 15:12'!
version
		"Allows polymoprhism with ClassDescription>>version"
		
	^ self traitVersion! !

Object subclass: #TraitComposition
	instanceVariableNames: 'transformations'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Composition'!

!TraitComposition commentStamp: '<historical>' prior: 0!
I hold a collection of trait transformations and provide important facilities to query the trait composition. For each trait in the composition clause there exists exactly one transformation in the collection.

Note, that directly manipulating the composition of a class or trait does not effect changes automatically. Use PureBehavior>>setTraitComposition: to do this. You have to make a copy of the old trait composition before changing it because only the difference between the new and the old composition is updated!!!

!TraitComposition class methodsFor: 'instance creation' stamp: 'al 12/12/2003 18:24'!
with: aTraitTransformation
	^self new
		add: aTraitTransformation;
		yourself! !

!TraitComposition class methodsFor: 'instance creation' stamp: 'al 12/12/2003 18:24'!
with: aTraitTransformation with: anotherTraitTransformation
	^self new
		add: aTraitTransformation;
		add: anotherTraitTransformation;
		yourself! !

!TraitComposition methodsFor: 'composition' stamp: 'apb 8/22/2005 17:52'!
+ aTraitExpression
	^ aTraitExpression addCompositionOnLeft: self.
! !

!TraitComposition methodsFor: 'composition' stamp: 'al 12/15/2003 20:56'!
- anArray
	"the modifier operators #@ and #- bind stronger than +.
	Thus, #@ or #- sent to a sum will only affect the most right summand"
	
	self transformations
		addLast: (self transformations removeLast - anArray)! !

!TraitComposition methodsFor: 'composition' stamp: 'al 12/15/2003 20:56'!
@ anArrayOfAssociations
	"the modifier operators #@ and #- bind stronger than +.
	Thus, #@ or #- sent to a sum will only affect the most right summand"

	self transformations
		addLast: (self transformations removeLast @ anArrayOfAssociations)! !

!TraitComposition methodsFor: 'accessing' stamp: 'al 3/18/2004 14:10'!
add: aTraitTransformation
	self errorIfNotAddable: aTraitTransformation.
	self transformations addLast: aTraitTransformation! !

!TraitComposition methodsFor: 'testing' stamp: 'stephane.ducasse 4/13/2009 20:33'!
addCompositionOnLeft: aTraitComposition
	self transformations do: [ :each | aTraitComposition add: each ].
	^ aTraitComposition! !

!TraitComposition methodsFor: 'composition' stamp: 'apb 8/22/2005 17:28'!
addOnTheLeft: aTrait 
	self errorIfNotAddable: aTrait.
	self transformations addFirst: aTrait! !

!TraitComposition methodsFor: 'accessing' stamp: 'al 3/26/2006 22:03'!
allTraits
	^self traits gather: [:trait |
		trait hasTraitComposition
			ifTrue: [trait traitComposition allTraits copyWith: trait]
			ifFalse: [Array with: trait]]! !

!TraitComposition methodsFor: 'converting' stamp: 'al 1/18/2004 19:35'!
asTraitComposition
	^self! !

!TraitComposition methodsFor: 'error-handling' stamp: 'al 3/18/2004 14:19'!
assertValidUser: aBehavior
	"Assert that this trait composition set for aBehavior
	does not introduce a cycle."
	
	(self allTraits includes: aBehavior) ifTrue: [
		TraitCompositionException signal: 'Cycle in compositions:  The composition (in)directly includes this trait!!']! !

!TraitComposition methodsFor: 'enquiries' stamp: 'nice 1/5/2010 15:59'!
changedSelectorsComparedTo: oldComposition
	| changedSelectors traits |
	changedSelectors := IdentitySet new.
	traits := self traits asIdentitySet addAll: oldComposition traits asIdentitySet; yourself.
	traits do: [:each | | oldTransformation newTransformation |
		newTransformation := self transformationOfTrait: each.
		oldTransformation := oldComposition transformationOfTrait: each.
		(newTransformation isNil or: [oldTransformation isNil])
			ifTrue: [
				changedSelectors addAll: each selectors]
			ifFalse: [
				changedSelectors addAll: 
					(newTransformation changedSelectorsComparedTo: oldTransformation)]].
	^changedSelectors! !

!TraitComposition methodsFor: 'copying' stamp: 'nice 1/13/2010 22:04'!
copy
	self error: 'should not be called'.
	^super copy! !

!TraitComposition methodsFor: 'copying' stamp: 'stephane.ducasse 4/13/2009 20:32'!
copyTraitExpression
	| newCopy |
	newCopy := self shallowCopy.
	newCopy transformations: (self transformations collect: [ :each | each copyTraitExpression ]).
	^ newCopy
! !

!TraitComposition methodsFor: 'copying' stamp: 'al 12/7/2005 20:49'!
copyWithExclusionOf: aSymbol to: aTrait
	| composition transformation |
	composition := self copyTraitExpression.
	transformation := (composition transformationOfTrait: aTrait).
	^composition
		remove: transformation;
		add: (transformation addExclusionOf: aSymbol);
		yourself! !

!TraitComposition methodsFor: 'copying' stamp: 'al 12/7/2005 20:50'!
copyWithoutAlias: aSymbol of: aTrait
	| composition transformation |
	composition := self copyTraitExpression.
	transformation := (composition transformationOfTrait: aTrait).
	^composition
		remove: transformation;
		add: (transformation removeAlias: aSymbol);
		normalizeTransformations;
		yourself! !

!TraitComposition methodsFor: 'error-handling' stamp: 'al 3/17/2004 15:22'!
errorIfNotAddable: aTraitTransformation
	(self includesTrait: aTraitTransformation trait) ifTrue: [
		^TraitCompositionException
			signal: 'Trait ' , aTraitTransformation trait asString, ' already in composition']! !

!TraitComposition methodsFor: 'private' stamp: 'al 7/30/2004 15:50'!
includesMethod: aSelector
	^(self methodDescriptionForSelector: aSelector) isEmpty not! !

!TraitComposition methodsFor: 'private' stamp: 'al 1/18/2004 19:32'!
includesTrait: aTrait
	^self traits includes: aTrait! !

!TraitComposition methodsFor: 'initialization' stamp: 'al 3/12/2004 17:06'!
initialize
	super initialize.
	transformations := OrderedCollection new! !

!TraitComposition methodsFor: 'private' stamp: 'al 4/10/2004 23:25'!
isAliasSelector: aSymbol
	"Return true if the selector aSymbol is an alias defined
	in this or in another composition somewhere deeper in 
	the tree of traits compositions."

	| methodDescription |
	methodDescription := (self methodDescriptionsForSelector: aSymbol)
		detect: [:each | each selector = aSymbol].
	^methodDescription isAliasSelector! !

!TraitComposition methodsFor: 'private' stamp: 'al 3/6/2004 19:05'!
isEmpty
	^self transformations isEmpty! !

!TraitComposition methodsFor: 'private' stamp: 'al 4/10/2004 23:34'!
isLocalAliasSelector: aSymbol
	"Return true if the selector aSymbol is an alias defined
	in this composition."

	| methodDescription |
	methodDescription := (self methodDescriptionsForSelector: aSymbol)
		detect: [:each | each selector = aSymbol].
	^methodDescription isLocalAliasSelector! !

!TraitComposition methodsFor: 'enquiries' stamp: 'al 7/21/2004 21:03'!
methodDescriptionForSelector: aSymbol
	"Return a TraitMethodDescription for the selector aSymbol."

	| description |
	description := TraitMethodDescription selector: aSymbol.
	self transformations do: [:each |
		each collectMethodsFor: aSymbol into: description].
	^description! !

!TraitComposition methodsFor: 'enquiries' stamp: 'al 7/21/2004 21:03'!
methodDescriptionsForSelector: aSymbol
	"Return a collection of TraitMethodDescriptions for aSymbol and all the 
	aliases of aSymbol."

	| selectors collection |
	selectors := IdentitySet with: aSymbol.
	self transformations do: [:each |
		selectors addAll: (each aliasesForSelector: aSymbol)].
	collection := OrderedCollection new: selectors size.
	selectors do: [:each |
		collection add: (self methodDescriptionForSelector: each)].
	^collection! !

!TraitComposition methodsFor: 'composition' stamp: 'al 4/10/2004 00:08'!
normalizeTransformations
	self transformations: (
		self transformations collect: [:each |
			each normalized])! !

!TraitComposition methodsFor: 'private' stamp: 'al 3/6/2004 19:05'!
notEmpty
	^self isEmpty not! !

!TraitComposition methodsFor: 'alias' stamp: 'EstebanLorenzano 1/15/2013 18:56'!
originSelectorOf: aSelector
 
	self isEmpty ifTrue: [ ^ aSelector ].
	self transformations do:[ :transformation | 
		transformation allAliasesDict 
			at: aSelector ifPresent: [:value | ^ value ]].
		
	^aSelector! !

!TraitComposition methodsFor: 'copying' stamp: 'nice 1/13/2010 21:53'!
postCopy
	super postCopy.
	transformations := transformations collect: [:each | each copy].! !

!TraitComposition methodsFor: 'printing' stamp: 'al 12/16/2003 22:06'!
printOn: aStream
	self transformations isEmptyOrNil
		ifFalse: [
			self transformations
				do: [:each | aStream print: each]
				separatedBy: [aStream nextPutAll: ' + '] ]
		ifTrue: [aStream nextPutAll: '{}']
		! !

!TraitComposition methodsFor: 'printing' stamp: 'al 12/16/2003 22:16'!
printString
	^String streamContents: [:stream |
		self printOn: stream]! !

!TraitComposition methodsFor: 'accessing' stamp: 'al 4/9/2004 17:12'!
remove: aTransformation
	self transformations
		remove: aTransformation! !

!TraitComposition methodsFor: 'accessing' stamp: 'al 4/9/2004 17:12'!
removeFromComposition: aTrait
	self remove:
		(self transformationOfTrait: aTrait)! !

!TraitComposition methodsFor: 'accessing' stamp: 'al 1/18/2004 19:29'!
size
	^transformations size! !

!TraitComposition methodsFor: 'enquiries' stamp: 'MarianoMartinezPeck 6/26/2012 10:39'!
traitProvidingSelector: aSymbol
	"Return the trait which originally provides the method aSymbol or return nil
	if trait composition does not provide this selector or there is a conflict.
	Take aliases into account. Return the trait which the aliased method is defined in."

	| methodDescription locatedMethod |
	methodDescription := self methodDescriptionForSelector: aSymbol.
	(methodDescription isProvided not or: [methodDescription isConflict])	
		ifTrue: [^nil].
	locatedMethod := methodDescription providedLocatedMethod.
	^locatedMethod methodClass traitOrClassOfSelector: locatedMethod selector! !

!TraitComposition methodsFor: 'accessing' stamp: 'al 4/13/2004 15:30'!
traits
	^self transformations collect: [:each |
		each trait]! !

!TraitComposition methodsFor: 'accessing' stamp: 'al 3/7/2004 22:31'!
transformationOfTrait: aTrait
	"Return the transformation which holds aTrait
	or nil if this composition doesn't include aTrait."
	
	^self transformations
		detect: [:each | each trait = aTrait]
		ifNone: [nil]! !

!TraitComposition methodsFor: 'accessing' stamp: 'al 3/17/2004 19:34'!
transformations
	^transformations! !

!TraitComposition methodsFor: 'testing' stamp: 'al 3/19/2004 10:30'!
transformations: aCollection
	transformations := aCollection! !

Object subclass: #TraitMethodDescription
	instanceVariableNames: 'selector locatedMethods'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Composition'!

!TraitMethodDescription commentStamp: '<historical>' prior: 0!
Used by TraitComposition to encapsulates a collection of methods for one particular selector when querying for changes. According to the number and kind of those methods a provided method exists, there is a conflict or there are no provided nor conflicting methods at all. I provide the interface to query for those situations, e.g., effectiveMethod returns the provided method or the conflict marker method.
!

!TraitMethodDescription class methodsFor: 'private' stamp: 'NS 4/19/2004 16:26'!
maxArguments
	^30! !

!TraitMethodDescription class methodsFor: 'instance creation' stamp: 'al 1/23/2004 12:52'!
new
	^super new
		initialize;
		yourself! !

!TraitMethodDescription class methodsFor: 'instance creation' stamp: 'al 1/23/2004 13:04'!
selector: aSymbol
	^self new
		selector: aSymbol
		yourself! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'al 1/23/2004 10:20'!
addLocatedMethod: aLocatedMethod
	locatedMethods add: aLocatedMethod! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'AdrianLienhard 10/18/2009 14:04'!
conflictMethod
	| method argumentNames binary numberOfArguments |
	self isConflict ifFalse: [^nil].
	argumentNames := self getArgumentNames.
	binary := self isBinarySelector.
	numberOfArguments := binary
		ifTrue: [1]
		ifFalse: [argumentNames size + 2].
	^self
		generateMethod: self selector
		withMarker: CompiledMethod conflictMarker
		forArgs: argumentNames size
		binary: binary! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'NS 4/19/2004 16:13'!
effectiveMethod
	"Return the effective compiled method of this method description." 

	| locatedMethod method |
	method := self providedMethod.
	method isNil ifFalse: [^ method].
	method := self conflictMethod.
	method isNil ifFalse: [^ method].
	^ self requiredMethod.! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'NS 4/16/2004 15:06'!
effectiveMethodCategory
	^ self effectiveMethodCategoryCurrent: nil new: nil! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'nice 1/5/2010 15:59'!
effectiveMethodCategoryCurrent: currentCategoryOrNil new: newCategoryOrNil
	| result size isCurrent isConflict |
	size := self size.
	size = 0 ifTrue: [^ nil].
	result := self locatedMethods anyOne category.
	size = 1 ifTrue: [^ result].
	
	isCurrent := currentCategoryOrNil isNil.
	isConflict := false.
	self locatedMethods do: [:each | | cat |
		cat := each category.
		isCurrent := isCurrent or: [cat == currentCategoryOrNil].
		isConflict := isConflict or: [cat ~~ result]].
	isConflict ifFalse: [^ result].
	(isCurrent not and: [newCategoryOrNil notNil]) ifTrue: [^ newCategoryOrNil].
	^ ClassOrganizer ambiguous.! !

!TraitMethodDescription methodsFor: 'private' stamp: 'mf 4/29/2012 13:15'!
generateMethod: aSelector withMarker: aSymbol forArgs: aNumber binary: aBoolean
	| source node |
	source := String streamContents: [:stream |
		aNumber < 1
			ifTrue: [stream nextPutAll: 'selector']
			ifFalse: [aBoolean
				ifTrue: [
					stream nextPutAll: '* anObject']
				ifFalse: [
					1 to: aNumber do: [:argumentNumber |
						stream
							nextPutAll: 'with:'; space;
							nextPutAll: 'arg'; nextPutAll: argumentNumber asString; space]]].
		stream cr; tab; nextPutAll: 'self '; nextPutAll: aSymbol].
	
	node := self class compilerClass new
		compile: source
		in: self class
		notifying: nil
		ifFail: [].
	^(node generateWithSource) selector: aSelector; yourself! !

!TraitMethodDescription methodsFor: 'private' stamp: 'al 3/12/2004 17:04'!
getArgumentNames
	| argumentNamesCollection names defaultName |
	defaultName := 'arg'.
	argumentNamesCollection := self locatedMethods
		collect: [:each | each argumentNames ].
	names := Array new: argumentNamesCollection anyOne size.
	argumentNamesCollection do: [:collection |
		1 to: names size do: [:index |
			(names at: index) isNil
				ifTrue: [names at: index put: (collection at: index)]
				ifFalse: [(names at: index) ~= (collection at: index)
					ifTrue: [names at: index put: defaultName, index asString]]]].
	^names
		! !

!TraitMethodDescription methodsFor: 'initialization' stamp: 'mada 5/5/2012 12:57'!
initialize
	super initialize.
	locatedMethods := IdentitySet new! !

!TraitMethodDescription methodsFor: 'testing' stamp: 'mada 5/5/2012 11:30'!
isAliasSelector
	"Return true if the selector is an alias (if it is different
	from the original selector) or already an aliased method
	in the original location (recursively search the compositions).
	Return false, if not or if we have a conflict."
	
	| locatedMethod |
	^self size = 1 and: [
		locatedMethod := self locatedMethods anyOne.
		(locatedMethod selector ~= self selector) or: [
			locatedMethod methodClass isAliasSelector: self selector]]! !

!TraitMethodDescription methodsFor: 'testing' stamp: 'al 2/20/2004 22:18'!
isBinarySelector
	^self locatedMethods anyOne
		isBinarySelector! !

!TraitMethodDescription methodsFor: 'testing' stamp: 'mada 5/5/2012 11:25'!
isConflict
	| count originMethodReferenciel |
	count := 0.

	self locatedMethods size > 1
		ifTrue: 
			["If they are more than 1 located method, then check whether these methods have the same origin"
			originMethodReferenciel := self locatedMethods anyOne  origin.
			(self locatedMethods allSatisfy: [:each | each origin == originMethodReferenciel])
				ifTrue: [ ^ false ]].

	self methodsDo: [:each |
		each isProvided ifTrue: [
			count := count + 1.
			count > 1 ifTrue: [^true]]].
	^false! !

!TraitMethodDescription methodsFor: 'testing' stamp: 'al 1/23/2004 10:31'!
isEmpty
	^self size = 0! !

!TraitMethodDescription methodsFor: 'testing' stamp: 'al 4/10/2004 23:34'!
isLocalAliasSelector
	"Return true if the selector is an alias (if it is different
	from the original selector). Return false, if not or if we
	have a conflict."
	
	^self size = 1 and: [
		(self locatedMethods anyOne selector ~= self selector)]! !

!TraitMethodDescription methodsFor: 'testing' stamp: 'NS 4/19/2004 16:10'!
isProvided
	^ self providedMethod notNil! !

!TraitMethodDescription methodsFor: 'testing' stamp: 'mada 5/5/2012 11:25'!
isRequired
	self isEmpty ifTrue: [^ false].
	^ self locatedMethods allSatisfy: [:each | each isRequired]! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'al 1/23/2004 10:20'!
locatedMethods
	^locatedMethods! !

!TraitMethodDescription methodsFor: 'enumeration' stamp: 'mada 5/5/2012 11:24'!
methodsDo: aBlock
	self locatedMethods do: [:each |
		aBlock value: each ]! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'mada 5/5/2012 11:24'!
providedLocatedMethod
	| locatedMethod aLocatedMethod refOrigin |
	locatedMethod := nil.
	self locatedMethods ifEmpty: [ ^ nil].

	self locatedMethods size > 1
		ifTrue: [ 	aLocatedMethod := self locatedMethods anyOne.
						refOrigin := (aLocatedMethod methodClass >> aLocatedMethod selector) origin.
						(self locatedMethods 
							allSatisfy: [:each | each origin == refOrigin])
						ifTrue: [^ aLocatedMethod].  ].
	
	self locatedMethods do: [:each |
		each isProvided ifTrue: [
			locatedMethod isNil ifFalse: [^nil].
			locatedMethod := each]].
	^locatedMethod! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'marcus.denker 11/10/2008 10:04'!
providedMethod
	^self providedLocatedMethod ifNotNil: [:locatedMethod | locatedMethod method]! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'mada 5/5/2012 11:22'!
requiredMethod
	| argumentNames numberOfArguments binary |
	self isRequired ifFalse: [^nil].
	self size = 1 ifTrue: [^self locatedMethods anyOne].

	argumentNames := self getArgumentNames.
	binary := self isBinarySelector.
	numberOfArguments := binary
		ifTrue: [1]
		ifFalse: [argumentNames size + 2].
	 ^self
		generateMethod: self selector
		withMarker: CompiledMethod implicitRequirementMarker
		forArgs: argumentNames size
		binary: binary.! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'al 1/23/2004 10:19'!
selector
	^selector! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'al 1/23/2004 10:19'!
selector: aSymbol 
	selector := aSymbol! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'al 1/23/2004 12:54'!
size
	^self locatedMethods size! !

Object subclass: #TraitTransformation
	uses: TComposingDescription
	instanceVariableNames: 'subject'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Composition'!

!TraitTransformation commentStamp: 'apb 8/25/2005 11:54' prior: 0!
A trait transformation is an instance of one of my concrete subclasses, TraitAlias or TraitExclusion. These represent a transformation of a trait, specified by the alias and exclusion operators. 

I define an instance variable named subject which holds the object that is transformed.  Thus, an alias transformation has as its subject a trait, and a trait exclusion has as its subject either a trait alias or a trait. Each of the concrete transformation classes implement the method allSelectors according to the transformation it represents. 

(There was formerly a subclass called TraitHolder, which was the identity transformation and which did not modify the trait.  This was clearly redundant, and was removed.)!

TraitTransformation class
	uses: TComposingDescription classTrait
	instanceVariableNames: ''!

TraitTransformation subclass: #TraitAlias
	instanceVariableNames: 'aliases'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Composition'!

!TraitAlias commentStamp: '<historical>' prior: 0!
See comment of my superclass TraitTransformation.!

!TraitAlias class methodsFor: 'instance creation' stamp: 'al 3/17/2004 15:22'!
assertValidAliasDefinition: anArrayOfAssociations
	"Throw an exceptions if the alias definition is not valid.
	It is expected to be a collection of associations and
	the number of arguments of the alias selector has to
	be the same as the original selector."

	((anArrayOfAssociations isKindOf: Collection) and: [
		anArrayOfAssociations allSatisfy: [:each |
			each isKindOf: Association]]) ifFalse: [
		TraitCompositionException signal: 'Invalid alias definition: Not a collection of associations.'].
	
	(anArrayOfAssociations allSatisfy: [:association |
		(association key numArgs = association value numArgs and: [
			(association key numArgs = -1) not])]) ifFalse: [
		TraitCompositionException signal: 'Invalid alias definition: Alias and original selector have to have the same number of arguments.']! !

!TraitAlias class methodsFor: 'instance creation' stamp: 'al 3/12/2004 16:24'!
with: aTraitComposition aliases: anArrayOfAssociations
	self assertValidAliasDefinition: anArrayOfAssociations.
	^self new
		subject: aTraitComposition;
		aliases: anArrayOfAssociations;
		yourself! !

!TraitAlias methodsFor: 'composition' stamp: 'al 3/3/2004 19:49'!
- anArrayOfSelectors
	^TraitExclusion
		with: self
		exclusions: anArrayOfSelectors! !

!TraitAlias methodsFor: 'enumeration' stamp: 'al 4/7/2004 21:14'!
aliasNamed: aSymbol ifAbsent: aBlock
	^self aliases
		detect: [:association |  association key = aSymbol]
		ifNone: aBlock! !

!TraitAlias methodsFor: 'accessing' stamp: 'al 3/12/2004 16:43'!
aliases
	"Collection of associations where key is the
	alias and value the original selector."
	
	^aliases! !

!TraitAlias methodsFor: 'accessing' stamp: 'al 3/17/2004 15:22'!
aliases: anArrayOfAssociations
	| newNames |
	newNames := (anArrayOfAssociations collect: [:each | each key]) asIdentitySet.
	newNames size < anArrayOfAssociations size ifTrue: [
		TraitCompositionException signal: 'Cannot use the same alias name twice'].
	anArrayOfAssociations do: [:each |
		(newNames includes: each value) ifTrue: [
			TraitCompositionException signal: 'Cannot define an alias for an alias']].
	aliases := anArrayOfAssociations! !

!TraitAlias methodsFor: 'enquiries' stamp: 'al 3/12/2004 17:10'!
aliasesForSelector: aSymbol
	| selectors |
	selectors := self aliases
		select: [:association | association value = aSymbol]
		thenCollect: [:association | association key].
	^(super aliasesForSelector: aSymbol)
		addAll: selectors;
		yourself
		 ! !

!TraitAlias methodsFor: 'enquiries' stamp: 'al 7/13/2004 16:53'!
allAliasesDict
	| dict |
	dict := super allAliasesDict.
	self aliases do: [:assoc |
		dict at: assoc key put: assoc value].
	^dict! !

!TraitAlias methodsFor: 'enquiries' stamp: 'al 3/7/2004 23:26'!
allSelectors
	^self subject allSelectors
		addAll: (self aliases collect: [:each | each key]) asSet;
		yourself! !

!TraitAlias methodsFor: 'enquiries' stamp: 'al 4/7/2004 21:16'!
collectMethodsFor: aSelector into: methodDescription
	| originalSelector association |
	self subject
		collectMethodsFor: aSelector
		into: methodDescription.			
	association := self aliasNamed: aSelector ifAbsent: [nil].
	association notNil ifTrue: [
		originalSelector := association value.
		self subject
			collectMethodsFor: originalSelector
			into: methodDescription]! !

!TraitAlias methodsFor: 'copying' stamp: 'dvf 8/25/2005 16:07'!
copyTraitExpression
	^super copyTraitExpression
		aliases: self aliases deepCopy;
		yourself! !

!TraitAlias methodsFor: 'testing' stamp: 'al 4/10/2004 23:57'!
isEmpty
	^self aliases isEmpty! !

!TraitAlias methodsFor: 'copying' stamp: 'nice 1/13/2010 22:02'!
postCopy
	super postCopy.
	aliases := self aliases copy! !

!TraitAlias methodsFor: 'printing' stamp: 'al 2/22/2004 21:45'!
printOn: aStream
	super printOn: aStream.
	aStream
		space;
		nextPut: $@;
		space;
		nextPut: ${.
	self aliases do: [:each | aStream print: each]
		separatedBy: [aStream nextPutAll: '. '].
	aStream nextPut: $}! !

!TraitAlias methodsFor: 'composition' stamp: 'al 4/10/2004 23:51'!
removeAlias: aSymbol
	self aliases: (self aliases
		reject: [:each | each key = aSymbol])! !

TraitTransformation subclass: #TraitExclusion
	instanceVariableNames: 'exclusions'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Composition'!

!TraitExclusion commentStamp: '<historical>' prior: 0!
See comment of my superclass TraitTransformation.!

!TraitExclusion class methodsFor: 'instance creation' stamp: 'dvf 8/18/2005 17:36'!
with: aTraitComposition exclusions: anArrayOfSelectors
	^self new
		subject: aTraitComposition;
		exclusions: anArrayOfSelectors;
		yourself
! !

!TraitExclusion methodsFor: 'composition' stamp: 'al 4/9/2004 16:55'!
addExclusionOf: aSymbol
	self exclusions: (self exclusions copyWith: aSymbol)! !

!TraitExclusion methodsFor: 'enquiries' stamp: 'al 3/7/2004 22:48'!
allSelectors
	| selectors |
	selectors := self subject allSelectors.
	self exclusions do: [:each |
		selectors remove: each ifAbsent: []].
	^selectors! !

!TraitExclusion methodsFor: 'enquiries' stamp: 'al 3/4/2004 16:35'!
collectMethodsFor: aSelector into: methodDescription
	(self exclusions includes: aSelector) ifFalse: [
		self subject
			collectMethodsFor: aSelector
			into: methodDescription]! !

!TraitExclusion methodsFor: 'copying' stamp: 'dvf 8/25/2005 16:07'!
copyTraitExpression
	^super copyTraitExpression 
		exclusions: self exclusions deepCopy;
		yourself! !

!TraitExclusion methodsFor: 'accessing' stamp: 'al 12/7/2003 11:58'!
exclusions
	^exclusions! !

!TraitExclusion methodsFor: 'accessing' stamp: 'al 12/7/2003 11:58'!
exclusions: aCollection
	exclusions := aCollection! !

!TraitExclusion methodsFor: 'testing' stamp: 'al 4/10/2004 23:56'!
isEmpty
	^self exclusions isEmpty! !

!TraitExclusion methodsFor: 'copying' stamp: 'nice 1/13/2010 21:52'!
postCopy
	super postCopy.
	exclusions := exclusions copy! !

!TraitExclusion methodsFor: 'printing' stamp: 'al 2/22/2004 21:45'!
printOn: aStream
	super printOn: aStream. 
	aStream
		space;
		nextPut: $-;
		space;
		nextPut: ${.
	self exclusions do: [:each | aStream print: each]
		separatedBy: [aStream nextPutAll: '. '].
	aStream nextPut: $}! !

!TraitTransformation methodsFor: 'composition' stamp: 'al 12/7/2005 21:19'!
- anArray
	TraitCompositionException signal: 'Invalid trait exclusion. Exclusions have to be specified after aliases.'! !

!TraitTransformation methodsFor: 'composition' stamp: 'al 12/7/2005 21:20'!
@ anArrayOfAssociations
	TraitCompositionException signal: 'Invalid trait exclusion. Aliases have to be specified before exclusions.'! !

!TraitTransformation methodsFor: 'composition' stamp: 'al 4/9/2004 17:01'!
addExclusionOf: aSymbol
	^self - {aSymbol}! !

!TraitTransformation methodsFor: 'enquiries' stamp: 'al 3/11/2004 21:49'!
aliasesForSelector: aSymbol
	"Return a collection of alias selectors that are defined in this transformation."
	
	^self subject aliasesForSelector: aSymbol! !

!TraitTransformation methodsFor: 'enquiries' stamp: 'NS 4/16/2004 13:40'!
allAliasesDict
	"Return a dictionary with all alias associations that are defined in this transformation."
	
	^self subject allAliasesDict! !

!TraitTransformation methodsFor: 'enquiries' stamp: 'al 3/7/2004 22:36'!
allSelectors
	^self subclassResponsibility! !

!TraitTransformation methodsFor: 'enquiries' stamp: 'NS 4/16/2004 14:20'!
changedSelectorsComparedTo: aTraitTransformation
	| selectors otherSelectors changedSelectors aliases otherAliases |
	selectors := self allSelectors asIdentitySet.
	otherSelectors := aTraitTransformation allSelectors asIdentitySet.
	changedSelectors := IdentitySet withAll: (
		(selectors difference: otherSelectors) union: (otherSelectors difference: selectors)).
	aliases := self allAliasesDict.
	otherAliases := aTraitTransformation allAliasesDict.
	aliases keysAndValuesDo: [:key :value |
		(value ~~ (otherAliases at: key ifAbsent: [nil])) ifTrue: [changedSelectors add: key]].
	otherAliases keysAndValuesDo: [:key :value |
		(value ~~ (aliases at: key ifAbsent: [nil])) ifTrue: [changedSelectors add: key]].
	^ changedSelectors.! !

!TraitTransformation methodsFor: 'enquiries' stamp: 'al 3/6/2004 18:51'!
collectMethodsFor: aSelector into: methodDescription
	"Collect instances of LocatedMethod into methodDescription
	for each method that has the selector aSelector and is not excluded
	or for which aSelector is an alias."

	self subclassResponsibility! !

!TraitTransformation methodsFor: 'copying' stamp: 'nice 1/13/2010 22:02'!
copy
	self error: 'should not be called'.
	^super copy! !

!TraitTransformation methodsFor: 'copying' stamp: 'dvf 8/25/2005 16:06'!
copyTraitExpression
	^self shallowCopy 
		subject: self subject copyTraitExpression;
		yourself! !

!TraitTransformation methodsFor: 'testing' stamp: 'al 4/10/2004 23:56'!
isEmpty
	self subclassResponsibility! !

!TraitTransformation methodsFor: 'accessing parallel hierarchy' stamp: 'dvf 8/18/2005 18:29'!
isMeta
	^self subject isMeta! !

!TraitTransformation methodsFor: 'accessing' stamp: 'al 4/10/2004 00:06'!
normalized
	^self isEmpty
		ifFalse: [
			self subject: self subject normalized.
			self]
		ifTrue: [self subject normalized]
		
		! !

!TraitTransformation methodsFor: 'copying' stamp: 'nice 1/13/2010 21:54'!
postCopy
	super postCopy.
	subject := subject copy! !

!TraitTransformation methodsFor: 'printing' stamp: 'al 12/12/2003 18:29'!
printOn: aStream
	aStream print: self subject! !

!TraitTransformation methodsFor: 'composition' stamp: 'al 4/10/2004 23:48'!
removeAlias: aSymbol
	self subject removeAlias: aSymbol! !

!TraitTransformation methodsFor: 'enquiries' stamp: 'dvf 8/18/2005 18:39'!
selectors
	^self allSelectors! !

!TraitTransformation methodsFor: 'browser support' stamp: 'dvf 8/18/2005 18:31'!
sourceCodeTemplate
	^ self subject sourceCodeTemplate! !

!TraitTransformation methodsFor: 'accessing' stamp: 'NS 12/11/2003 15:13'!
subject
	^subject! !

!TraitTransformation methodsFor: 'accessing' stamp: 'al 4/9/2004 18:23'!
subject: aTraitTransformation
	subject := aTraitTransformation! !

!TraitTransformation methodsFor: 'accessing parallel hierarchy' stamp: 'dvf 8/18/2005 18:28'!
theNonMetaClass 
	^ self subject theNonMetaClass  ! !

!TraitTransformation methodsFor: 'enquiries' stamp: 'al 12/12/2003 18:00'!
trait
	^self subject trait! !

!TraitTransformation methodsFor: 'enquiries' stamp: 'dvf 8/16/2005 17:35'!
traitTransformations 
	^ { subject }! !

Error subclass: #TraitException
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Kernel'!

!TraitException commentStamp: '<historical>' prior: 0!
General exception used for example to signal invalid trait compositions!

TraitException subclass: #TraitCompositionException
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Composition'!

!TraitCompositionException commentStamp: '<historical>' prior: 0!
Signal invalid trait compositions.!
